#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import copy
import imp

try:
    from unittest import mock
except ImportError:
    import mock

from octavia.common import data_models as o_data_models

from a10_octavia.common.data_models import VThunder
import a10_octavia.controller.worker.tasks.nat_pool_tasks as task
from a10_octavia.tests.common import a10constants
from a10_octavia.tests.unit import base

VTHUNDER = VThunder()
LB = o_data_models.LoadBalancer(id=a10constants.MOCK_LOAD_BALANCER_ID)
NAT_POOL_FLAVOR1 = {u'nat_pool': {u'start_address': u'172.16.1.101',
                                  u'end_address': u'172.16.1.108',
                                  u'netmask': u'/24', u'gateway': u'172.16.1.1',
                                  u'pool_name': u'pool1'}}
FLAVOR1 = {u'start_address': u'172.16.2.101',
           u'end_address': u'172.16.2.102', u'netmask': u'/24',
           u'gateway': u'172.16.2.1', u'pool_name': u'pooln1'}
FLAVOR2 = {u'start_address': u'172.16.3.101',
           u'end_address': u'172.16.3.102', u'netmask': u'/24',
           u'gateway': u'172.16.3.1', u'pool_name': u'pooln2'}
NAT_POOL_FLAVOR2 = {u'nat_pool_list': [FLAVOR1, FLAVOR2]}


class TestHandlerNatPoolTasks(base.BaseTaskTestCase):

    def setUp(self):
        super(TestHandlerNatPoolTasks, self).setUp()
        imp.reload(task)
        self.client_mock = mock.Mock()

    def tearDown(self):
        super(TestHandlerNatPoolTasks, self).tearDown()

    def test_create_nat_pool_with_nat_pool_not_exists_with_nat_pool_flavor(self):
        vthunder = copy.deepcopy(VTHUNDER)
        mock_nat_pool_task = task.NatPoolCreate()
        mock_nat_pool_task.axapi_client = self.client_mock
        self.client_mock.nat.pool.exists.return_value = False
        self.client_mock.nat.pool.try_get.return_value = None
        mock_nat_pool_task.execute(LB, vthunder, flavor_data=NAT_POOL_FLAVOR1)
        args, kwargs = self.client_mock.nat.pool.create.call_args
        self.assertEqual(kwargs['pool_name'], 'pool1')
        self.assertEqual(kwargs['start_address'], '172.16.1.101')
        self.assertEqual(kwargs['gateway'], '172.16.1.1')

    def test_create_nat_pool_with_nat_pool_exists_with_nat_pool_flavor(self):
        device_pool = {u'pool': {u'uuid': u'947c228e-443a-11eb-9273-525400bf462d',
                                 u'start-address': u'172.16.1.101', u'port-overload': 0,
                                 u'a10-url': u'/axapi/v3/ip/nat/pool/pool1', u'netmask': u'/24',
                                 u'end-address': u'172.16.1.108', u'ip-rr': 0,
                                 u'gateway': u'172.16.1.1', u'pool-name': u'pool1'}}

        vthunder = copy.deepcopy(VTHUNDER)
        mock_nat_pool_task = task.NatPoolCreate()
        mock_nat_pool_task.axapi_client = self.client_mock
        self.client_mock.nat.pool.exists.return_value = True
        self.client_mock.nat.pool.try_get.return_value = device_pool
        mock_nat_pool_task.execute(LB, vthunder, flavor_data=NAT_POOL_FLAVOR1)
        self.client_mock.nat.pool.create.not_called()

    def test_create_nat_pool_with_nat_pool_not_exists_with_nat_pool_list_flavor(self):
        vthunder = copy.deepcopy(VTHUNDER)
        mock_nat_pool_task = task.NatPoolCreate()
        mock_nat_pool_task.axapi_client = self.client_mock
        self.client_mock.nat.pool.exists.return_value = False
        self.client_mock.nat.pool.try_get.return_value = None
        mock_nat_pool_task.execute(LB, vthunder, flavor_data=NAT_POOL_FLAVOR2)
        args, kwargs = self.client_mock.nat.pool.create.call_args
        self.assertEqual(kwargs['pool_name'], 'pooln2')
        self.assertEqual(kwargs['gateway'], '172.16.3.1')

    def test_create_nat_pool_with_nat_pool_exists_with_nat_pool_list_flavor(self):
        device_pool = {u'pool': {u'uuid': u'eef8b1ea-48dc-11eb-b050-5254000780f9',
                                 u'start-address': u'172.16.2.101', u'port-overload': 0,
                                 u'a10-url': u'/axapi/v3/ip/nat/pool/pooln1', u'netmask': u'/24',
                                 u'end-address': u'172.16.2.102', u'ip-rr': 0,
                                 u'gateway': u'172.16.2.1', u'pool-name': u'pooln1'}}

        vthunder = copy.deepcopy(VTHUNDER)
        mock_nat_pool_task = task.NatPoolCreate()
        mock_nat_pool_task.axapi_client = self.client_mock
        self.client_mock.nat.pool.exists.return_value = True
        self.client_mock.nat.pool.try_get.return_value = device_pool
        mock_nat_pool_task.execute(LB, vthunder, flavor_data=NAT_POOL_FLAVOR2)
        self.client_mock.nat.pool.create.not_called()

    def test_delete_nat_pool_when_only_one_virtual_server_uses_it(self):
        vthunder = copy.deepcopy(VTHUNDER)
        mock_nat_pool_task = task.NatPoolDelete()
        lb_count = 1
        mock_nat_pool_task.axapi_client = self.client_mock
        mock_nat_pool_task.execute(LB, vthunder, lb_count, flavor_data=NAT_POOL_FLAVOR1)
        args, kwargs = self.client_mock.nat.pool.delete.call_args
        self.assertIn('pool1', args)

    def test_delete_nat_pool_when_multiple_virtual_servers_use_it(self):
        vthunder = copy.deepcopy(VTHUNDER)
        mock_nat_pool_task = task.NatPoolDelete()
        lb_count = 2
        mock_nat_pool_task.axapi_client = self.client_mock
        task_path = "a10_octavia.controller.worker.tasks.nat_pool_tasks"
        log_message = str("Cannot delete Nat-pool(s) in flavor None as they are in use by "
                          "another loadbalancer(s)")
        expected_log = ["WARNING:{}:{}".format(task_path, log_message)]
        with self.assertLogs(task_path, level='WARN') as cm:
            mock_nat_pool_task.execute(LB, vthunder, lb_count, flavor_data=NAT_POOL_FLAVOR1)
            self.assertEqual(expected_log, cm.output)

    def test_delete_nat_pool_when_virtual_port_uses_it(self):
        vthunder = copy.deepcopy(VTHUNDER)
        mock_nat_pool_task = task.NatPoolDelete()
        lb_count = 1
        mock_nat_pool_task.axapi_client = self.client_mock
        mock_nat_pool_task.execute(LB, vthunder, lb_count, flavor_data=NAT_POOL_FLAVOR1)
        self.client_mock.nat.pool.delete.not_called()



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from acos_client import errors as acos_errors
from oslo_config import cfg
from oslo_log import log as logging
from requests.exceptions import ConnectionError
from taskflow import task

from octavia.common import exceptions

from a10_octavia.common import a10constants
from a10_octavia.common import openstack_mappings
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator_for_revert
from a10_octavia.controller.worker.tasks import utils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def _get_hm_name(axapi_client, health_mon):
    try:
        hm = axapi_client.slb.hm.get(health_mon.id)
    except (acos_errors.NotFound):
        # Backwards compatability with a10-neutron-lbaas
        hm = axapi_client.slb.hm.get(health_mon.id[0:28])
    return hm['monitor']['name']


class CreateAndAssociateHealthMonitor(task.Task):
    """Task to create a healthmonitor and associate it with provided pool."""

    @axapi_client_decorator
    def execute(self, listeners, health_mon, vthunder, flavor=None):
        method = None
        url = None
        expect_code = None

        if health_mon.type in a10constants.HTTP_TYPE:
            method = health_mon.http_method
            url = health_mon.url_path
            expect_code = health_mon.expected_codes
        args = utils.meta(health_mon, 'hm', {})
        args = utils.dash_to_underscore(args)

        # overwrite options from flavor
        if flavor:
            flavors = flavor.get('health_monitor')
            if flavors:
                name_exprs = flavors.get('name_expressions')
                parsed_exprs = utils.parse_name_expressions(health_mon.name, name_exprs)
                flavors.pop('name_expressions', None)
                flavors.update(parsed_exprs)
                args.update({'monitor': flavors})

        try:
            health_mon.type = openstack_mappings.hm_type(self.axapi_client, health_mon.type)
        except Exception:
            raise exceptions.ProviderUnsupportedOptionError(
                prov="A10",
                user_msg=("Failed to create health monitor {}, "
                          "A health monitor of type {} is not supported "
                          "by A10 provider").format(health_mon.id, health_mon.type))

        try:
            post_data = CONF.health_monitor.post_data
            self.axapi_client.slb.hm.create(health_mon.id,
                                            health_mon.type,
                                            health_mon.delay, health_mon.timeout,
                                            health_mon.rise_threshold, method=method,
                                            port=listeners[0].protocol_port, url=url,
                                            expect_code=expect_code, post_data=post_data,
                                            **args)
            LOG.debug("Successfully created health monitor: %s", health_mon.id)

        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception("Failed to create health monitor: %s", health_mon.id)
            raise e

        try:
            self.axapi_client.slb.service_group.update(health_mon.pool_id,
                                                       hm_name=health_mon.id,
                                                       health_check_disable=0)
            LOG.debug("Successfully associated health monitor %s to pool %s",
                      health_mon.id, health_mon.pool_id)
        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception(
                "Failed to associate health monitor %s to pool %s",
                health_mon.id, health_mon.pool_id)
            raise e

    @axapi_client_decorator_for_revert
    def revert(self, listeners, health_mon, vthunder, *args, **kwargs):
        try:
            self.axapi_client.slb.hm.delete(health_mon.id)
        except ConnectionError:
            LOG.exception("Failed to connect A10 Thunder device: %s", vthunder.ip_address)
        except Exception as e:
            LOG.warning(
                "Failed to revert creation of health monitor: %s due to %s",
                health_mon.id, str(e))


class DeleteHealthMonitor(task.Task):
    """Task to disassociate Health Monitor from pool and delete"""

    @axapi_client_decorator
    def execute(self, health_mon, vthunder):
        try:
            self.axapi_client.slb.service_group.update(health_mon.pool_id,
                                                       hm_delete=True)
            LOG.debug("Successfully dissociated health monitor %s from pool %s",
                      health_mon.id, health_mon.pool_id)
        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception(
                "Failed to dissociate health monitor %s from pool %s",
                health_mon.pool_id, health_mon.id)
            raise e

        try:
            hm_name = _get_hm_name(self.axapi_client, health_mon)
            self.axapi_client.slb.hm.delete(hm_name)
            LOG.debug("Successfully deleted health monitor: %s", health_mon.id)
        except acos_errors.NotFound:
            LOG.debug("Health monitor %s was already deleted. Skipping...", health_mon.id)
        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception("Failed to delete health monitor: %s", health_mon.id)
            raise e


class UpdateHealthMonitor(task.Task):
    """Task to update Health Monitor"""

    @axapi_client_decorator
    def execute(self, listeners, health_mon, vthunder, update_dict, flavor=None):
        """ Execute update health monitor """

        health_mon.update(update_dict)
        method = None
        url = None
        expect_code = None
        if health_mon.type in a10constants.HTTP_TYPE:
            method = health_mon.http_method
            url = health_mon.url_path
            expect_code = health_mon.expected_codes
        args = utils.meta(health_mon, 'hm', {})
        args = utils.dash_to_underscore(args)

        # overwrite options from flavor
        if flavor:
            flavors = flavor.get('health_monitor')
            if flavors:
                name_exprs = flavors.get('name_expressions')
                parsed_exprs = utils.parse_name_expressions(health_mon.name, name_exprs)
                flavors.pop('name_expressions', None)
                flavors.update(parsed_exprs)
                args.update({'monitor': flavors})

        try:
            hm_name = _get_hm_name(self.axapi_client, health_mon)
            post_data = CONF.health_monitor.post_data
            self.axapi_client.slb.hm.update(
                hm_name,
                openstack_mappings.hm_type(self.axapi_client, health_mon.type),
                health_mon.delay, health_mon.timeout, health_mon.rise_threshold,
                method=method, url=url, expect_code=expect_code, post_data=post_data,
                port=listeners[0].protocol_port, **args)
            LOG.debug("Successfully updated health monitor: %s", health_mon.id)
        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception("Failed to update health monitor: %s", health_mon.id)
            raise e



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from octavia.common import data_models as o_data_models
from octavia.tests.unit import base

from a10_octavia.controller.worker.tasks import policy
from a10_octavia.tests.common import a10constants

L7POLICY = o_data_models.L7Policy(id=a10constants.MOCK_L7POLICY_ID,
                                  listener_id=a10constants.MOCK_LISTENER_ID,
                                  action="REDIRECT_TO_URL",
                                  redirect_url="www.google.com")
L7RULE = o_data_models.L7Rule(id=a10constants.MOCK_L7RULE_ID,
                              l7policy_id=a10constants.MOCK_L7POLICY_ID,
                              type="PATH", compare_type="EQUAL_TO",
                              value="www://abc.com")


class TestPolicy(base.TestCase):
    def test_policy_rule_path(self):
        L7POLICY.l7rules = [L7RULE]
        policy_util = policy.PolicyUtil()
        policy_util.createPolicy(L7POLICY)
        self.assertEqual(policy_util.ruleParser(L7RULE),
                         '([HTTP::path] equals "www://abc.com")')

    def test_policy_rule_file_type_regex(self):
        L7rule_file_type = o_data_models.L7Rule(id=a10constants.MOCK_L7RULE_ID,
                                                l7policy_id=a10constants.MOCK_L7POLICY_ID,
                                                type="FILE_TYPE", compare_type="REGEX",
                                                value=".(html|xml)")
        L7POLICY.l7rules = [L7rule_file_type]
        policy_util = policy.PolicyUtil()
        self.assertEqual(policy_util.ruleParser(L7rule_file_type),
                         '([HTTP::uri] matches_regex ".(html|xml)")')

    def test_policy_rule_file_type_equals(self):
        L7rule_file_type = o_data_models.L7Rule(id=a10constants.MOCK_L7RULE_ID,
                                                l7policy_id=a10constants.MOCK_L7POLICY_ID,
                                                type="FILE_TYPE", compare_type="EQUAL_TO",
                                                value="ccccc")
        L7POLICY.l7rules = [L7rule_file_type]
        policy_util = policy.PolicyUtil()
        self.assertEqual(policy_util.ruleParser(L7rule_file_type),
                         '([HTTP::uri] ends_with "ccccc")')



"""Added defaults for non-nullable tables

Revision ID: a3c90b57952d
Revises: 34a0d811941d
Create Date: 2021-03-31 23:10:50.297427

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = 'a3c90b57952d'
down_revision = '34a0d811941d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('nat_pool', 'member_ref_count',
               existing_type=mysql.INTEGER(display_width=11),
               nullable=False)
    op.alter_column('vrid', 'vrid',
               existing_type=mysql.INTEGER(display_width=11),
               nullable=True)
    op.alter_column('vrid', 'vrid_floating_ip',
               existing_type=mysql.VARCHAR(length=40),
               nullable=True)
    op.alter_column('vthunders', 'hierarchical_multitenancy',
               existing_type=mysql.VARCHAR(length=7),
               nullable=False)
    op.alter_column('vthunders', 'last_udp_update',
               existing_type=mysql.DATETIME(),
               nullable=False)
    op.alter_column('vthunders', 'partition_name',
               existing_type=mysql.VARCHAR(length=14),
               nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('vthunders', 'partition_name',
               existing_type=mysql.VARCHAR(length=14),
               nullable=True)
    op.alter_column('vthunders', 'last_udp_update',
               existing_type=mysql.DATETIME(),
               nullable=True)
    op.alter_column('vthunders', 'hierarchical_multitenancy',
               existing_type=mysql.VARCHAR(length=7),
               nullable=True)
    op.alter_column('vrid', 'vrid_floating_ip',
               existing_type=mysql.VARCHAR(length=40),
               nullable=False)
    op.alter_column('vrid', 'vrid',
               existing_type=mysql.INTEGER(display_width=11),
               nullable=False)
    op.alter_column('nat_pool', 'member_ref_count',
               existing_type=mysql.INTEGER(display_width=11),
               nullable=True)
    # ### end Alembic commands ###



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import datetime

from oslo_config import cfg
from oslo_log import log as logging
from sqlalchemy.orm import noload
from sqlalchemy import or_
from sqlalchemy import and_

from octavia.common import constants as consts
from octavia.db import models as base_models
from octavia.db import repositories as repo

from a10_octavia.db import models

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class BaseRepository(object):
    model_class = None

    def count(self, session, **filters):
        """Retrieves a count of entities from the database.

        :param session: A Sql Alchemy database session.
        :param filters: Filters to decide which entities should be retrieved.
        :returns: int
        """
        return session.query(self.model_class).filter_by(**filters).count()

    def create(self, session, **model_kwargs):
        """Base create method for a database entity.

        :param session: A Sql Alchemy database session.
        :param model_kwargs: Attributes of the model to insert.
        :returns: octavia.common.data_model
        """
        with session.begin(subtransactions=True):
            model = self.model_class(**model_kwargs)
            session.add(model)
        return model.to_data_model()

    def delete(self, session, **filters):
        """Deletes entities from the database.

        :param session: A Sql Alchemy database session.
        :param filters: Filters to decide which entity should be deleted.
        :returns: None
        :raises: sqlalchemy.orm.exc.NoResultFound
        """
        models = session.query(self.model_class).filter_by(**filters).all()
        for model in models:
            with session.begin(subtransactions=True):
                session.delete(model)
                session.flush()

    def delete_batch(self, session, ids=None):
        """Batch deletes by entity ids."""
        ids = ids or []
        for id in ids:
            self.delete(session, id=id)

    def update(self, session, id, **model_kwargs):
        """Updates an entity in the database.

        :param session: A Sql Alchemy database session.
        :param model_kwargs: Entity attributes that should be updates.
        :returns: octavia.common.data_model
        """
        with session.begin(subtransactions=True):
            session.query(self.model_class).filter_by(
                id=id).update(model_kwargs)

    def get(self, session, **filters):
        """Retrieves an entity from the database.

        :param session: A Sql Alchemy database session.
        :param filters: Filters to decide which entity should be retrieved.
        :returns: octavia.common.data_model
        """
        deleted = filters.pop('show_deleted', True)
        model = session.query(self.model_class).filter_by(**filters)

        if not deleted:
            if hasattr(self.model_class, 'status'):
                model = model.filter(
                    self.model_class.status != consts.DELETED)
            else:
                model = model.filter(
                    self.model_class.provisioning_status != consts.DELETED)

        model = model.first()

        if not model:
            return None

        return model.to_data_model()

    def get_all(self, session, pagination_helper=None,
                query_options=None, **filters):
        """Retrieves a list of entities from the database.

        :param session: A Sql Alchemy database session.
        :param pagination_helper: Helper to apply pagination and sorting.
        :param query_options: Optional query options to apply.
        :param filters: Filters to decide which entities should be retrieved.
        :returns: [octavia.common.data_model]
        """
        deleted = filters.pop('show_deleted', True)
        query = session.query(self.model_class).filter_by(**filters)
        if query_options:
            query = query.options(query_options)

        if not deleted:
            if hasattr(self.model_class, 'status'):
                query = query.filter(
                    self.model_class.status != consts.DELETED)
            else:
                query = query.filter(
                    self.model_class.provisioning_status != consts.DELETED)

        if pagination_helper:
            model_list, links = pagination_helper.apply(
                query, self.model_class)
        else:
            links = None
            model_list = query.all()

        data_model_list = [model.to_data_model() for model in model_list]
        return data_model_list, links

    def exists(self, session, id):
        """Determines whether an entity exists in the database by its id.

        :param session: A Sql Alchemy database session.
        :param id: id of entity to check for existence.
        :returns: octavia.common.data_model
        """
        return bool(session.query(self.model_class).filter_by(id=id).first())

    def get_all_deleted_expiring(self, session, exp_age):
        """Get all previously deleted resources that are now expiring.

        :param session: A Sql Alchemy database session.
        :param exp_age: A standard datetime delta which is used to see for how
                        long can a resource live without updates before
                        it is considered expired
        :returns: A list of resource IDs
                """

        expiry_time = datetime.datetime.utcnow() - exp_age

        query = session.query(self.model_class).filter(
            self.model_class.updated_at < expiry_time)
        if hasattr(self.model_class, 'status'):
            query = query.filter_by(status=consts.DELETED)
        else:
            query = query.filter_by(provisioning_status=consts.DELETED)
        # Do not load any relationship
        query = query.options(noload('*'))
        model_list = query.all()

        id_list = [model.id for model in model_list]
        return id_list


class VThunderRepository(BaseRepository):
    model_class = models.VThunder

    def get_recently_updated_thunders(self, session):
        query = session.query(self.model_class).filter(
            or_(self.model_class.updated_at >= self.model_class.last_write_mem,
                self.model_class.last_write_mem == None)).filter(
            or_(self.model_class.role == "STANDALONE",
                self.model_class.role == "MASTER")).filter(
            or_(self.model_class.status == 'ACTIVE', self.model_class.status == 'DELETED'))
        query = query.options(noload('*'))
        return query.all()

    def get_stale_vthunders(self, session, initial_setup_wait_time, failover_wait_time):
        model = session.query(self.model_class).filter(
            self.model_class.created_at < initial_setup_wait_time).filter(
            self.model_class.last_udp_update < failover_wait_time).filter(
            or_(self.model_class.status == 'ACTIVE',
                self.model_class.status == 'READY')).filter(
            self.model_class.health_state == 'UP').first()
        if model is None:
            return None
        return model.to_data_model()

    def get_compute_vthunders(self, session, compute_id):
        vthunder_list = []
        query = session.query(self.model_class).filter(
            self.model_class.compute_id == compute_id)
        model_list = query.all()
        for data in model_list:
            vthunder_list.append(data.to_data_model())
        return vthunder_list

    def set_vthunder_health_state(self, session, id, new_state):
        self.update(session, id, health_state=new_state)

    def unset_vthunder_busy_health_state(self, session):
        with session.begin(subtransactions=True):
            session.query(self.model_class).filter_by(
                health_state='BUSY').update({"health_state": 'UP'})

    def get_vthunder_from_lb(self, session, lb_id):
        model = session.query(self.model_class).filter(
            self.model_class.loadbalancer_id == lb_id).filter(
            or_(self.model_class.role == "STANDALONE",
                self.model_class.role == "MASTER")).first()

        if not model:
            return None

        return model.to_data_model()

    def get_backup_vthunder_from_lb(self, session, lb_id):
        model = session.query(self.model_class).filter(
            self.model_class.loadbalancer_id == lb_id).filter(
            or_(self.model_class.role == "STANDALONE",
                self.model_class.role == "BACKUP")).first()

        if not model:
            return None

        return model.to_data_model()

    def get_health_vthunder_count_for_lb(self, session, lb_id):
        count = session.query(self.model_class).filter(
            self.model_class.loadbalancer_id == lb_id).filter(
            self.model_class.health_state == "UP").count()
        return count

    def get_vthunder_by_project_id(self, session, project_id):
        model = session.query(self.model_class).filter(
            self.model_class.project_id == project_id).filter(
            and_(self.model_class.status == "ACTIVE",
                 or_(self.model_class.role == "STANDALONE",
                     self.model_class.role == "MASTER"))).first()

        if not model:
            return None

        return model.to_data_model()

    def get_vthunder_with_different_topology(self, session, project_id, topology):
        model = session.query(self.model_class).filter(
            self.model_class.project_id == project_id).filter(
            and_(self.model_class.status == "ACTIVE",
                 or_(self.model_class.role == "STANDALONE",
                     self.model_class.role == "MASTER"))).filter(
                self.model_class.compute_id != None).filter(
                self.model_class.topology != topology).first()

        if not model:
            return None

        return model.to_data_model()

    def get_vthunders_by_project_id(self, session, project_id):
        model_list = session.query(self.model_class).filter(
            self.model_class.project_id == project_id).filter(
            and_(self.model_class.status == "ACTIVE",
                 or_(self.model_class.role == "STANDALONE",
                     self.model_class.role == "MASTER")))
        id_list = [model.id for model in model_list]
        return id_list

    def get_vthunders_by_ip_address(self, session, ip_address, vthunders=False):
        model_list = session.query(self.model_class).filter(
            self.model_class.ip_address == ip_address).filter(
            and_(self.model_class.status == "ACTIVE",
                 or_(self.model_class.role == "STANDALONE",
                     self.model_class.role == "MASTER")))

        if vthunders == False:
            id_list = [model.id for model in model_list]
            return id_list
        else:
            model_list = model_list.options(noload('*'))
            return model_list.all()

    def get_all_vthunder_by_address(self, session, ip_address):
        model_list = session.query(self.model_class).filter(
            self.model_class.ip_address == ip_address).filter(
            self.model_class.status == "ACTIVE")
        model_list = model_list.options(noload('*'))
        return model_list.all()

    def get_delete_compute_flag(self, session, compute_id):
        if compute_id:
            count = session.query(self.model_class).filter(
                and_(self.model_class.compute_id == compute_id,
                     self.model_class.status != "DELETED")).count()
            if count < 2:
                return True

            else:
                return False
        else:
            return False

    def get_vthunder_from_src_addr(self, session, srcaddr):
        model = session.query(self.model_class).filter(
            self.model_class.status != "DELETED").filter(
            self.model_class.ip_address == srcaddr).first()

        if not model:
            return None
        return model.id

    def get_spare_vthunder(self, session):
        model = session.query(self.model_class).filter(
            self.model_class.status == "READY").first()

        if not model:
            return None

        return model.to_data_model()

    def set_spare_vthunder_status(self, session, id, new_status):
        self.update(session, id, status=new_status)

    def get_spare_vthunder_count(self, session):
        with session.begin(subtransactions=True):
            count = session.query(self.model_class).filter_by(
                status="READY", loadbalancer_id=None).count()

        return count

    def get_busy_spare_vthunder_count(self, session):
        with session.begin(subtransactions=True):
            count = session.query(self.model_class).filter_by(
                status="BUSY", loadbalancer_id=None).count()

        return count

    def get_all_deleted_expiring(self, session, exp_age):

        expiry_time = datetime.datetime.utcnow() - exp_age

        query = session.query(self.model_class).filter(
            self.model_class.updated_at < expiry_time)
        if hasattr(self.model_class, 'status'):
            query = query.filter(or_(self.model_class.status == "USED_SPARE",
                                     self.model_class.status == consts.DELETED))
        else:
            query = query.filter_by(operating_status=consts.DELETED)
        # Do not load any relationship
        query = query.options(noload('*'))
        model_list = query.all()

        id_list = [model.id for model in model_list]
        return id_list

    def get_project_list_using_partition(self, session, partition_name, ip_address):
        queryset_vthunders = session.query(self.model_class.project_id.distinct()).filter(
            and_(self.model_class.partition_name == partition_name,
                 self.model_class.ip_address == ip_address,
                 or_(self.model_class.role == "STANDALONE",
                     self.model_class.role == "MASTER",
                     self.model_class.role == "BACKUP")))
        list_projects = [project[0] for project in queryset_vthunders]
        return list_projects

    def update_last_write_mem(self, session, ip_address, partition, **model_kwargs):
        with session.begin(subtransactions=True):
            session.query(self.model_class).filter_by(
                    ip_address=ip_address, partition_name=partition).update(model_kwargs)

    def get_vthunder_by_project_id_and_role(self, session, project_id, role):
        model = session.query(self.model_class).filter(
            self.model_class.project_id == project_id).filter(
            and_(self.model_class.status == "ACTIVE",
                 self.model_class.role == role)).first()

        if not model:
            return None

        return model.to_data_model()

    def get_vthunders_by_project_id_and_role(self, session, project_id, role):
        model_list = session.query(self.model_class).filter(
            self.model_class.project_id == project_id).filter(
            and_(self.model_class.status == "ACTIVE",
                 self.model_class.role == role))

        id_list = [model.id for model in model_list]
        return id_list

    def get_rack_vthunders_by_ip_address(self, session, ip_address):
        model_list = session.query(self.model_class).filter(
            self.model_class.ip_address == ip_address).filter(
            and_(self.model_class.role == "MASTER",
                 or_(self.model_class.status == "ACTIVE",
                     self.model_class.status == "PENDING_DELETE")))

        id_list = [model.id for model in model_list]
        return id_list

    def get_lb_count_vthunder_partition(self, session, ip_address, partition):
        status_list = ["ACTIVE", "PENDING_UPDATE"]
        return session.query(self.model_class).filter(self.model_class.ip_address == ip_address,
                                                    self.model_class.partition_name == partition,
                                                    self.model_class.status.in_(status_list)).count()


class LoadBalancerRepository(repo.LoadBalancerRepository):
    thunder_model_class = models.VThunder

    def get_active_lbs_by_thunder(self, session, vthunder):
        lb_list = []
        query = session.query(self.model_class).filter(
            and_(vthunder.loadbalancer_id == self.model_class.id,
                 self.model_class.provisioning_status == consts.ACTIVE))
        model_list = query.all()
        for data in model_list:
            lb_list.append(data.to_data_model())
        return lb_list

    def get_lb_count_by_subnet(self, session, project_ids, subnet_id):
        return session.query(self.model_class).join(base_models.Vip).filter(
            and_(self.model_class.project_id.in_(project_ids),
                 base_models.Vip.subnet_id == subnet_id,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE,
                     self.model_class.provisioning_status == consts.PENDING_UPDATE))).count()

    def get_lb_count_by_flavor(self, session, project_ids, flavor_id):
        return session.query(self.model_class).filter(
            self.model_class.project_id.in_(project_ids)).filter(
                self.model_class.flavor_id == flavor_id,
                or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                    self.model_class.provisioning_status == consts.ACTIVE)).count()

    def get_lbs_by_subnet(self,session, project_ids, subnet_id):
        lb_list = []
        query = session.query(self.model_class).join(base_models.Vip).filter(
            and_(self.model_class.project_id.in_(project_ids),
                 base_models.Vip.subnet_id == subnet_id,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE,
                     self.model_class.provisioning_status == consts.PENDING_UPDATE)))
        model_list = query.all()
        for data in model_list:
            lb_list.append(data.to_data_model())
        return lb_list

    def check_lb_exists_in_project(self, session, project_id):
        lb = session.query(self.model_class).join(base_models.Vip).filter(
            and_(self.model_class.project_id == project_id,
                 self.model_class.provisioning_status == consts.ACTIVE)).count()
        if lb == 0:
            return False
        else:
            return True

    def get_lbs_by_project_id(self, session, project_id):
        lb_list = []
        query = session.query(self.model_class).filter(
            self.model_class.project_id == project_id).filter(
            or_(self.model_class.provisioning_status == "ACTIVE",
                self.model_class.provisioning_status == "PENDING_UPDATE",
                self.model_class.provisioning_status == "PENDING_CREATE"))

        model_list = query.all()
        for data in model_list:
            lb_list.append(data.to_data_model())
        return lb_list

    def get_all_other_lbs_in_project(self, session, project_id, id):
        lb_list = []
        query = session.query(self.model_class).filter(
            self.model_class.project_id == project_id).filter(
            and_(self.model_class.id != id,
                 self.model_class.provisioning_status != "ERROR",
                 self.model_class.provisioning_status != "DELETED"))

        model_list = query.all()
        for data in model_list:
            lb_list.append(data.to_data_model())
        return lb_list

    def get_lb_excluding_deleted(self, session, lb_id):
        query = session.query(self.model_class).filter(
            and_(self.model_class.id == lb_id,
                 self.model_class.provisioning_status != "DELETED"))
        model = query.first()
        if model is None:
            return None
        return model.to_data_model()

    def get_lbs_on_thunder_by_subnet(self, session, loadbalancer_id, subnet_id):
        model = session.query(self.model_class).join(base_models.Vip).filter(
            and_(self.model_class.id == loadbalancer_id,
                 base_models.Vip.subnet_id == subnet_id,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE,
                     self.model_class.provisioning_status == consts.PENDING_UPDATE))).first()
        if not model:
            return None
        return model.to_data_model()


class VRIDRepository(BaseRepository):
    model_class = models.VRID

    # A project can have multiple VRIDs, so need to convert each vrid object through
    # "to_data_model"
    def get_vrid_from_owner(self, session, owner, project_ids):
        vrid_obj_list = []

        model = session.query(self.model_class).filter(
            or_(self.model_class.owner.in_(owner),
                self.model_class.owner.in_(project_ids)))
        for data in model:
            vrid_obj_list.append(data.to_data_model())

        return vrid_obj_list


class MemberRepository(repo.MemberRepository):

    def get_member_count(self, session, project_id):
        count = session.query(self.model_class).filter(
            and_(self.model_class.project_id == project_id,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE))).count()
        return count

    def get_member_count_by_subnet(self, session, project_ids, subnet_id):
        return session.query(self.model_class).filter(
            and_(self.model_class.project_id.in_(project_ids),
                 self.model_class.subnet_id == subnet_id,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE))).count()

    def get_member_count_by_ip_address(self, session, ip_address, project_id):
        return session.query(self.model_class).filter(
            self.model_class.project_id == project_id).filter(
            and_(self.model_class.ip_address == ip_address,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE))).count()

    def get_member_count_by_ip_address_port_protocol(self, session, ip_address, project_id, port,
                                                     protocol):
        return session.query(self.model_class).join(base_models.Pool).filter(
            self.model_class.project_id == project_id).filter(
            and_(self.model_class.ip_address == ip_address,
                 self.model_class.protocol_port == port,
                 base_models.Pool.protocol == protocol,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE))).count()

    def get_pool_count_by_ip(self, session, ip_address, project_id):
        return session.query(self.model_class.pool_id.distinct()).filter(
            self.model_class.project_id == project_id).filter(
            and_(self.model_class.ip_address == ip_address,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE))).count()

    def get_pool_count_by_ip_on_thunder(self, session, ip_address, pool_ids):
        return session.query(self.model_class.pool_id.distinct()).filter(
            self.model_class.pool_id.in_(pool_ids)).filter(
            and_(self.model_class.ip_address == ip_address,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE))).count()

    def get_pool_count_subnet(self, session, project_ids, subnet_id):
        return session.query(self.model_class.pool_id.distinct()).filter(
            self.model_class.project_id.in_(project_ids)).filter(
            and_(self.model_class.subnet_id == subnet_id,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE))).count()

    def get_pool_count_subnet_on_thunder(self, session, pool_ids, subnet_id):
        return session.query(self.model_class.pool_id.distinct()).filter(
            self.model_class.pool_id.in_(pool_ids)).filter(
            and_(self.model_class.subnet_id == subnet_id,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE))).count()

    def get_members_on_thunder_by_subnet(self, session, ip_address, subnet_id):
        model = session.query(self.model_class).filter(
            and_(self.model_class.ip_address == ip_address,
                 self.model_class.subnet_id == subnet_id,
                 or_(self.model_class.provisioning_status == consts.PENDING_DELETE,
                     self.model_class.provisioning_status == consts.ACTIVE))).first()

        if not model:
            return None
        return model.to_data_model()

    def get_members_by_project_id(self, session, project_id):
        member_list = []
        query = session.query(self.model_class).filter(
            self.model_class.project_id == project_id).filter(
            or_(self.model_class.provisioning_status == "ACTIVE",
                self.model_class.provisioning_status == "PENDING_UPDATE",
                self.model_class.provisioning_status == "PENDING_CREATE"))

        model_list = query.all()
        for data in model_list:
            member_list.append(data.to_data_model())
        return member_list


class NatPoolRepository(BaseRepository):
    model_class = models.NATPool


class VrrpSetRepository(BaseRepository):
    model_class = models.VrrpSet

    def get_subnet_set_ids(self, session, subnet):
        query = session.query(self.model_class).filter(
            self.model_class.mgmt_subnet == subnet)
        query = query.options(noload('*'))
        model_list = query.all()
        set_id_list = [model.set_id for model in model_list]
        return set_id_list

class ListenerStatisticsRepository(repo.ListenerStatisticsRepository):

    def delete_multiple(self, session, **filters):
        """Deletes entities from the database."""

        models = session.query(self.model_class).filter_by(**filters).all()
        for model in models:
            with session.begin(subtransactions=True):
                session.delete(model)
                session.flush()



[DEFAULT]
DEFAULT_VTHUNDER_USERNAME = "admin"
DEFAULT_VTHUNDER_PASSWORD = "a10"
DEFAULT_AXAPI_VERSION = 30

[SLB]

arp_disable = False
default_virtual_server_vrid = "10"
logging_template = "Logging_temp1"
policy_template = "policy_temp1"
template_virtual_server = "virtual_server_template1"
default_virtual_server_vrid = 0


[LISTENER]
ipinip = False
no_dest_nat = False
ha_conn_mirror = False
template_virtual_port = "vport_template"
template_tcp = "tcp_template"
template_policy = "policy_temp1"
autosnat = True
conn_limit = 5000
template_http = "http_template"
use_rcv_hop_for_resp = True

[SERVICE - GROUP]
templates = "server1"

[L7POLICY]
P1 = "P1"
P2 = "P2"

[L7RULE]
R1 = "R1"
R2 = "R2"

[SERVER]
conn_limit = 5000
conn_resume = 1
templates = "server1"

[RACK_VTHUNDER]
devices = """[
                    {
                     "project_id":"a0e57f9fcdfa47d18fe9ec9f80d63851",
                     "ip_address":"10.43.82.155",
                     "undercloud":"True",
                     "username":"admin",
                     "password":"a10",
                     "device_name":"rack_vthunder",
                     "axapi_version":"30"
                     },
                     {
                     "project_id":"a0e37f9fcdfa48d18fe9ec9f80d6385f",
                     "ip_address":"10.43.12.137",
                     "undercloud":"True",
                     "username":"admin",
                     "password":"a10",
                     "device_name":"rack_vthunder",
                     "axapi_version":"30",
                     "role":"MASTER",
                     "topology":"STANDALONE"
                     }
             ]
       """



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

try:
    import mock
except ImportError:
    from unittest import mock

from neutronclient.common import exceptions as neutron_client_exceptions

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

from octavia.common import clients
from octavia.network import base as network_driver_base
from octavia.network.drivers.neutron import allowed_address_pairs
from octavia.network.drivers.neutron import base as neutron_base
from octavia.tests.common import data_model_helpers as dmh
from octavia.tests.unit import base

from a10_octavia.common import a10constants
from a10_octavia.common import config_options
from a10_octavia.network.drivers.neutron import a10_octavia_neutron
from a10_octavia.tests.common import a10constants as a10_tconstants


class TestA10NeutronDriver(base.TestCase):

    def setUp(self):
        super(TestA10NeutronDriver, self).setUp()
        with mock.patch('octavia.common.clients.neutron_client.Client',
                        autospec=True) as neutron_client:
            with mock.patch('stevedore.driver.DriverManager.driver',
                            autospec=True):
                client = neutron_client(clients.NEUTRON_VERSION)
                client.list_extensions.return_value = {
                    'extensions': [
                        {'alias': allowed_address_pairs.AAP_EXT_ALIAS},
                        {'alias': neutron_base.SEC_GRP_EXT_ALIAS}
                    ]
                }
                self.k_session = mock.patch(
                    'keystoneauth1.session.Session').start()
                self.driver = a10_octavia_neutron.A10OctaviaNeutronDriver()
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        self.conf.register_opts(config_options.A10_CONTROLLER_WORKER_OPTS,
                                group=a10_tconstants.A10_CONTROLLER_WORKER_CONF_SECTION)

    def _setup_deallocate_vip_mocks(self):
        self.driver._get_lb_security_group = mock.MagicMock()
        self.driver._get_ports_by_security_group = mock.MagicMock()
        self.driver._get_instance_ports_by_subnet = mock.MagicMock()
        self.driver._remove_security_group = mock.MagicMock()
        self.driver._cleanup_port = mock.MagicMock()
        self.driver._delete_vip_security_group = mock.MagicMock()

    def _generate_lb(self):
        lb = dmh.generate_load_balancer_tree()
        lb.vip.load_balancer = lb
        return lb

    def test_deallocate_vip_one_lb(self):
        lb = self._generate_lb()
        sec_group = {'id': 'sec-grp-1'}
        ports = [{'id': 'instance-port-1'},
                 {'id': 'instance-port-2'},
                 {'id': lb.vip.port_id}]

        self._setup_deallocate_vip_mocks()
        get_lb_sec = self.driver._get_lb_security_group
        get_ports_sec = self.driver._get_ports_by_security_group
        cleanup_port = self.driver._cleanup_port
        delete_sec_grp = self.driver._delete_vip_security_group
        get_lb_sec.return_value = sec_group
        get_ports_sec.return_value = ports

        calls = []
        for port in ports:
            calls.append(mock.call(lb.vip.port_id, port))
        self.driver.deallocate_vip(lb, 1)
        cleanup_port.assert_has_calls(calls, any_order=True)
        delete_sec_grp.assert_called_once_with(sec_group['id'])

    def test_deallocate_vip_multi_lb(self):
        lb = self._generate_lb()
        sec_group = {'id': 'sec-grp-1'}
        ports = [{'id': 'instance-port-1'},
                 {'id': 'instance-port-2'},
                 {'id': lb.vip.port_id}]

        self._setup_deallocate_vip_mocks()
        get_lb_sec = self.driver._get_lb_security_group
        get_ports_sec = self.driver._get_ports_by_security_group
        rem_sec_grp = self.driver._remove_security_group
        cleanup_port = self.driver._cleanup_port
        delete_sec_grp = self.driver._delete_vip_security_group
        get_lb_sec.return_value = sec_group
        get_ports_sec.return_value = ports

        rem_calls = []
        for port in ports:
            rem_calls.append(mock.call(port, sec_group['id']))
        self.driver.deallocate_vip(lb, 2)
        rem_sec_grp.assert_has_calls(rem_calls, any_order=True)
        cleanup_port.assert_called_once_with(lb.vip.port_id, {'id': lb.vip.port_id})
        delete_sec_grp.assert_called_once_with(sec_group['id'])

    def test_deallocate_vip_sec_grp_disabled(self):
        lb = self._generate_lb()
        vip_port = {'id': lb.vip.port_id}
        ports = [{'id': 'instance-port-1'},
                 {'id': 'instance-port-2'}]

        self.driver.sec_grp_enabled = False
        self._setup_deallocate_vip_mocks()
        self.driver.neutron_client.show_port.return_value = vip_port
        get_ports_subnet = self.driver._get_instance_ports_by_subnet
        cleanup_port = self.driver._cleanup_port
        delete_sec_grp = self.driver._delete_vip_security_group
        get_ports_subnet.return_value = ports

        calls = []
        for port in ports:
            calls.append(mock.call(lb.vip.port_id, port))
        self.driver.deallocate_vip(lb, 1)
        cleanup_port.assert_has_calls(calls, any_order=True)
        delete_sec_grp.assert_not_called()

    def test_deallocate_vip_no_sec_grp_ports(self):
        lb = self._generate_lb()
        sec_group = {'id': 'sec-grp-1'}
        vip_port = {'id': lb.vip.port_id}
        ports = [{'id': 'instance-port-1'},
                 {'id': 'instance-port-2'}]

        self._setup_deallocate_vip_mocks()
        self.driver.neutron_client.show_port.return_value = vip_port
        get_lb_sec = self.driver._get_lb_security_group
        get_ports_sec = self.driver._get_ports_by_security_group
        get_ports_subnet = self.driver._get_instance_ports_by_subnet
        cleanup_port = self.driver._cleanup_port
        delete_sec_grp = self.driver._delete_vip_security_group
        get_lb_sec.return_value = sec_group
        get_ports_sec.return_value = []
        get_ports_subnet.return_value = ports

        calls = []
        for port in ports:
            calls.append(mock.call(lb.vip.port_id, port))
        self.driver.deallocate_vip(lb, 1)
        cleanup_port.assert_has_calls(calls, any_order=True)
        delete_sec_grp.assert_called_once_with(sec_group['id'])

    def test_deallocate_vip_no_sec_group(self):
        lb = self._generate_lb()
        vip_port = {'id': lb.vip.port_id}
        ports = [{'id': 'instance-port-1'},
                 {'id': 'instance-port-2'}]

        self._setup_deallocate_vip_mocks()
        self.driver.neutron_client.show_port.return_value = vip_port
        get_lb_sec = self.driver._get_lb_security_group
        get_ports_subnet = self.driver._get_instance_ports_by_subnet
        cleanup_port = self.driver._cleanup_port
        delete_sec_grp = self.driver._delete_vip_security_group
        rem_sec_grp = self.driver._remove_security_group
        get_lb_sec.return_value = None
        get_ports_subnet.return_value = ports

        self.driver.deallocate_vip(lb, 2)
        cleanup_port.assert_called_with(lb.vip.port_id, vip_port)
        delete_sec_grp.assert_not_called()
        rem_sec_grp.assert_not_called()

    def test_get_ports_by_security_group(self):
        ports = {'ports': [{'id': 'instance-port-1',
                            'security_groups': ['sec-grp-1', 'sec-grp-2'],
                            'device_owner': a10constants.OCTAVIA_OWNER},
                           {'id': 'instance-port-2',
                            'security_groups': ['sec-grp-1', 'sec-grp-2'],
                            'device_owner': a10constants.OCTAVIA_OWNER},
                           {'id': 'vrrp-port-1',
                            'security_groups': ['sec-grp-1'],
                            'device_owner': a10constants.OCTAVIA_OWNER}]}
        self.driver.project_id = 'proj-1'
        self.driver.neutron_client.list_ports.return_value = ports

        actual_ports = self.driver._get_ports_by_security_group('sec-grp-1')
        self.assertEqual(ports['ports'], actual_ports)

    def test_get_ports_by_security_group_empty_ret(self):
        ports = {}
        self.driver.project_id = 'proj-1'
        self.driver.neutron_client.list_ports.return_value = ports

        actual_ports = self.driver._get_ports_by_security_group('sec-grp-1')
        self.assertEqual([], actual_ports)

    def test_get_ports_by_security_group_empty_secgrps(self):
        ports = {'ports': [{'id': 'instance-port-1',
                            'security_groups': [],
                            'device_owner': a10constants.OCTAVIA_OWNER},
                           {'id': 'instance-port-2',
                            'security_groups': [],
                            'device_owner': a10constants.OCTAVIA_OWNER},
                           {'id': 'vrrp-port-1',
                            'security_groups': [],
                            'device_owner': a10constants.OCTAVIA_OWNER}]}
        self.driver.project_id = 'proj-1'
        self.driver.neutron_client.list_ports.return_value = ports

        actual_ports = self.driver._get_ports_by_security_group('sec-grp-1')
        self.assertEqual([], actual_ports)

    def test_get_ports_by_security_group_oct_not_owner(self):
        ports = {'ports': [{'id': 'instance-port-1',
                            'security_groups': ['sec-grp-1', 'sec-grp-2'],
                            'device_owner': 'outis'},
                           {'id': 'instance-port-2',
                            'security_groups': ['sec-grp-1', 'sec-grp-2'],
                            'device_owner': 'outis'},
                           {'id': 'vrrp-port-1',
                            'security_groups': ['sec-grp-1'],
                            'device_owner': 'outis'}]}
        self.driver.project_id = 'proj-1'
        self.driver.neutron_client.list_ports.return_value = ports

        actual_ports = self.driver._get_ports_by_security_group('sec-grp-1')
        self.assertEqual(ports['ports'], actual_ports)

    def test_get_instance_ports_by_subnet(self):
        ports = {'ports': [{'id': 'instance-port-1',
                            'fixed_ips': [{'subnet_id': 'subnet-1'}],
                            'device_owner': a10constants.OCTAVIA_OWNER},
                           {'id': 'instance-port-2',
                            'fixed_ips': [{'subnet_id': 'subnet-1'}],
                            'device_owner': a10constants.OCTAVIA_OWNER},
                           {'id': 'instance-port-3',
                            'fixed_ips': [{'subnet_id': 'subnet-20'}],
                            'device_owner': a10constants.OCTAVIA_OWNER}]}
        self.driver.neutron_client.list_ports.return_value = ports

        actual_ports = self.driver._get_instance_ports_by_subnet('amp-comp-1', 'subnet-1')
        ports['ports'].pop(2)
        self.assertEqual(ports['ports'], actual_ports)

    def test_get_instance_ports_by_subnet_oct_not_owner(self):
        ports = {'ports': [{'id': 'instance-port-1',
                            'fixed_ips': [{'subnet_id': 'subnet-1'}],
                            'device_owner': 'outis'},
                           {'id': 'instance-port-2',
                            'fixed_ips': [{'subnet_id': 'subnet-1'}],
                            'device_owner': 'outis'}]}
        self.driver.neutron_client.list_ports.return_value = ports

        actual_ports = self.driver._get_instance_ports_by_subnet('amp-comp-1', 'subnet-1')
        self.assertEqual([], actual_ports)

    def test_cleanup_port(self):
        vip_port_id = 'vrrp-port-1'
        port = {'id': 'instance-port-1'}
        delete_port = self.driver.neutron_client.delete_port
        self.driver._cleanup_port(vip_port_id, port)
        delete_port.assert_called_with(port['id'])

    def test_cleanup_port_vip_port_deleted(self):
        vip_port_id = 'vrrp-port-1'
        port = {'id': vip_port_id}
        delete_port = self.driver.neutron_client.delete_port
        delete_port.side_effect = neutron_client_exceptions.NotFound

        log_path = "a10_octavia.network.drivers.neutron.a10_octavia_neutron"
        log_message = str("VIP port %s already deleted. Skipping.")
        expected_log = ["DEBUG:{}:{}".format(
            log_path, log_message % vip_port_id)]

        with self.assertLogs(log_path, level='DEBUG') as cm:
            self.driver._cleanup_port(vip_port_id, port)
            delete_port.assert_called_with(vip_port_id)
            self.assertEqual(expected_log, cm.output)

    def test_cleanup_port_not_found(self):
        vip_port_id = 'vrrp-port-1'
        port = {'id': 'instance-port-1'}
        delete_port = self.driver.neutron_client.delete_port
        delete_port.side_effect = neutron_client_exceptions.PortNotFoundClient

        log_path = "a10_octavia.network.drivers.neutron.a10_octavia_neutron"
        log_message = str("Can't deallocate instance port %s because it "
                          "cannot be found in neutron. "
                          "Continuing cleanup.")
        expected_log = ["WARNING:{}:{}".format(
            log_path, log_message % port['id'])]

        with self.assertLogs(log_path, level='WARN') as cm:
            self.driver._cleanup_port(vip_port_id, port)
            delete_port.assert_called_with(port['id'])
            self.assertEqual(expected_log, cm.output)

    def test_cleanup_port_unknown_failure(self):
        vip_port_id = 'vrrp-port-1'
        port = {'id': 'instance-port-1'}
        delete_port = self.driver.neutron_client.delete_port
        delete_port.side_effect = Exception

        log_path = "a10_octavia.network.drivers.neutron.a10_octavia_neutron"
        log_message = ('Error deleting VIP port_id %s from neutron')
        expected_log = ["ERROR:{}:{}".format(
            log_path, log_message % port['id'])]

        expected_error = network_driver_base.DeallocateVIPException
        module_func = self.driver._cleanup_port
        module_args = [vip_port_id, port]
        with self.assertLogs(log_path, level='ERROR') as cm:
            self.assertRaises(expected_error, module_func, *module_args)
            delete_port.assert_called_with(port['id'])
            self.assertIn(expected_log[0], cm.output[0])

    def test_remove_security_group(self):
        sec_grp_id = 'sec-grp-1'
        port = {'id': 'instance-port-1',
                'security_groups': [sec_grp_id, 'sec-grp-2']}
        expected_payload = {'port': {'security_groups': ['sec-grp-2']}}
        update_port = self.driver.neutron_client.update_port

        self.driver._remove_security_group(port, sec_grp_id)
        update_port.assert_called_with(port['id'], expected_payload)

    def test_remove_security_group_not_found(self):
        sec_grp_id = 'sec-grp-1'
        port = {'id': 'instance-port-1',
                'security_groups': [sec_grp_id, 'sec-grp-2']}
        expected_payload = {'port': {'security_groups': ['sec-grp-2']}}
        update_port = self.driver.neutron_client.update_port
        update_port.side_effect = neutron_client_exceptions.PortNotFoundClient

        expected_error = network_driver_base.PortNotFound
        module_func = self.driver._remove_security_group
        module_args = [port, sec_grp_id]

        self.assertRaises(expected_error, module_func, *module_args)
        update_port.assert_called_with(port['id'], expected_payload)

    def test_remove_security_group_unknown_failure(self):
        sec_grp_id = 'sec-grp-1'
        port = {'id': 'instance-port-1',
                'security_groups': [sec_grp_id, 'sec-grp-2']}
        expected_payload = {'port': {'security_groups': ['sec-grp-2']}}
        update_port = self.driver.neutron_client.update_port
        update_port.side_effect = Exception

        expected_error = network_driver_base.NetworkException
        module_func = self.driver._remove_security_group
        module_args = [port, sec_grp_id]

        self.assertRaises(expected_error, module_func, *module_args)
        update_port.assert_called_with(port['id'], expected_payload)



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import copy
import imp
import json
try:
    from unittest import mock
except ImportError:
    import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

from octavia.common import data_models as o_data_models
from octavia.network import data_models as n_data_models
from octavia.network.drivers.neutron import utils

from acos_client import errors as acos_errors

from a10_octavia.common import config_options
from a10_octavia.common import data_models as a10_data_models
from a10_octavia.common import exceptions
from a10_octavia.common import utils as a10_utils
from a10_octavia.controller.worker.tasks import vthunder_tasks as task
from a10_octavia.network.drivers.neutron import a10_octavia_neutron
from a10_octavia.tests.common import a10constants
from a10_octavia.tests.unit import base

PROJECT_ID = "project-rack-vthunder"
VTHUNDER = a10_data_models.VThunder(
    project_id=PROJECT_ID, acos_version="5.2.1")
VLAN_ID = '11'
VE_IP_VLAN_ID = '12'
DELETE_VLAN = True
TAG_INTERFACE = '5'
TAG_TRUNK_INTERFACE = '1'
ETH_DATA = {"action": "enable"}
SUBNET_MASK = ["10.0.11.0", "255.255.255.0", "10.0.11.0/24"]
VE_IP_SUBNET_MASK = ["10.0.12.0", "255.255.255.0", "10.0.12.0/24"]
ETHERNET_INTERFACE = {
    "interface_num": 5,
    "vlan_map": [
        {"vlan_id": 11, "use_dhcp": "True"},
        {"vlan_id": 12, "ve_ip": ".10"}
    ]
}
TRUNK_INTERFACE = {
    "interface_num": 1,
    "vlan_map": [
        {"vlan_id": 12, "ve_ip": ".10"}
    ]
}
DEVICE1_MGMT_IP = "10.0.0.1"
DEVICE2_MGMT_IP = "10.0.0.2"
RACK_DEVICE = {
    "project_id": PROJECT_ID,
    "ip_address": "10.0.0.10",
    "device_name": "rack_vthunder",
    "username": "abc",
    "password": "abc",
    "interface_vlan_map": {
        "device_1": {
            "vcs_device_id": 1,
            "mgmt_ip_address": DEVICE1_MGMT_IP,
            "trunk_interfaces": [TRUNK_INTERFACE]
        }
    }
}
RACK_DEVICE_VCS = {
    "project_id": PROJECT_ID,
    "ip_address": "10.0.0.10",
    "device_name": "rack_vthunder",
    "username": "abc",
    "password": "abc",
    "interface_vlan_map": {
        "device_1": {
            "vcs_device_id": 1,
            "mgmt_ip_address": DEVICE1_MGMT_IP,
            "ethernet_interfaces": [ETHERNET_INTERFACE]
        },
        "device_2": {
            "vcs_device_id": 2,
            "mgmt_ip_address": DEVICE2_MGMT_IP,
            "ethernet_interfaces": [ETHERNET_INTERFACE]
        }
    }
}
DEL_VS_LIST = {"virtual-server-list": [{"ip-address": "10.0.1.1"}]}
DEL_SERVER_LIST = {"server-list": [{"host": "10.0.2.1"}]}
DEL_PORT_ID = "mock-port-id"
VE_IP = "10.0.11.10"
STATIC_VE_IP = "10.0.12.10"
NETWORK_11 = n_data_models.Network(id="mock-network-1", subnets=["mock-subnet-1"],
                                   provider_segmentation_id=11)
NETWORK_12 = n_data_models.Network(id="mock-network-2", subnets=["mock-subnet-2"],
                                   provider_segmentation_id=12)
VCS_DISABLED = {"vcs-summary": {"oper": {"vcs-enabled": "Invalid"}}}
VCS_MASTER_VBLADE = {
    "vcs-summary": {
        "oper": {
            "vcs-enabled": "Yes",
            "member-list": [{
                "port": 41216, "priority": 125, "state": "vMaster(*)",
                "ip-list": [{"ip": "10.0.0.1"}], "location": "Local", "id": 1
            }, {
                "port": 41216, "priority": 120, "state": "vBlade",
                "ip-list": [{"ip": "10.0.0.2"}], "location": "Remote", "id": 2
            }]
        }
    }
}
VCS_DEVICE1_FAILED = {
    "vcs-summary": {
        "oper": {
            "vcs-enabled": "Yes",
            "member-list": [{
                "port": 41216, "priority": 120, "state": "Unknown",
                "ip-list": [{"ip": "N/A"}], "location": "Remote", "id": 1
            }, {
                "port": 41216, "priority": 125, "state": "vMaster(*)",
                "ip-list": [{"ip": "10.0.0.2"}], "location": "Local", "id": 2
            }]
        }
    }
}
INTERFACES = {
    "interface": {
        "management": {
            "ip": {"dhcp": 1, "control-apps-use-mgmt-port": 0},
            "broadcast-rate-limit": {u'bcast-rate-limit-enable': 0},
            "flow-control": 0,
            "uuid": "6d4f3b4e-8beb-11eb-957b-fa163e54fddf",
            "a10-url": "/axapi/v3/interface/management"},
        "available-eth-list": {
            "uuid": "6d4f40c6-8beb-11eb-957b-fa163e54fddf",
            "a10-url": "/axapi/v3/interface/available-eth-list"},
        "brief": {
            "uuid": "6d4f42b0-8beb-11eb-957b-fa163e54fddf",
            "a10-url": "/axapi/v3/interface/brief"},
        "ethernet-list": [{
            "port-scan-detection": "disable",
            "uuid": "6d4f3fb8-8beb-11eb-957b-fa163e54fddf", "trap-source": 0,
            "ping-sweep-detection": "disable", "mtu": 1500, "ifnum": 1,
            "load-interval": 300, "a10-url": "/axapi/v3/interface/ethernet/1",
            "action": "disable", "virtual-wire": 0, "l3-vlan-fwd-disable": 0}],
        "a10-url": "/axapi/v3/interface",
        "common": {
            "uuid": "6d4f4224-8beb-11eb-957b-fa163e54fddf",
            "a10-url": "/axapi/v3/interface/common"}
    }
}

FLAVOR_DEV1 = a10_data_models.HardwareThunder(device_name="rack_thunder_1",
                                              undercloud=True, username="abc", password="abc",
                                              ip_address="10.10.10.10", partition_name="shared",
                                              device_name_as_key=True)
FLAVOR_DEV2 = a10_data_models.HardwareThunder(project_id="project-2", device_name="rack_thunder_2",
                                              undercloud=True, username="abc", password="abc",
                                              ip_address="10.10.10.10", partition_name="shared")
FLAVOR_DEV3 = a10_data_models.HardwareThunder(project_id="project-1", device_name="rack_thunder_1",
                                              undercloud=True, username="abc", password="abc",
                                              ip_address="10.10.10.10", partition_name="shared")
DEV_CONF_DICT = {'[dev]rack_thunder_1': FLAVOR_DEV1,
                 'project-1': FLAVOR_DEV3,
                 '[dev]rack_thunder_2': FLAVOR_DEV2,
                 'project-2': FLAVOR_DEV2}
DEV_FLAVOR = {'device_name': 'rack_thunder_1'}

VIP = o_data_models.Vip(ip_address="1.1.1.1")
AMPHORA = o_data_models.Amphora(id=a10constants.MOCK_AMPHORA_ID)
AMPHORAE = [AMPHORA]
LB = o_data_models.LoadBalancer(
    id=a10constants.MOCK_LOAD_BALANCER_ID, vip=VIP,
    amphorae=AMPHORAE, project_id=PROJECT_ID)
DEPLOYMENT_FLAVOR = {'deployment': {"dsr_type": "l2dsr_transparent"}}
SUBNET = n_data_models.Subnet()


class TestVThunderTasks(base.BaseTaskTestCase):

    def setUp(self):
        super(TestVThunderTasks, self).setUp()
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        self.conf.register_opts(config_options.A10_GLOBAL_OPTS,
                                group=a10constants.A10_GLOBAL_CONF_SECTION)
        imp.reload(task)
        self.client_mock = mock.Mock()
        self.db_session = mock.patch(
            'a10_octavia.controller.worker.tasks.a10_database_tasks.db_apis.get_session')
        self.db_session.start()

    def tearDown(self):
        super(TestVThunderTasks, self).tearDown()
        self.conf.reset()
        self.db_session.stop()

    def _mock_lb(self):
        lb = mock.Mock()
        lb.id = a10constants.MOCK_LOAD_BALANCER_ID
        lb.project_id = "project-rack-vthunder"
        lb.vip.subnet_id = "mock-subnet-1"
        return lb

    def _mock_member(self):
        member = mock.Mock()
        member.id = a10constants.MOCK_MEMBER_ID
        member.project_id = "project-rack-vthunder"
        member.subnet_id = "mock-subnet-1"
        return member

    @mock.patch('keystoneauth1.session.Session', mock.Mock())
    @mock.patch('neutronclient.client.SessionClient', mock.Mock())
    @mock.patch('neutronclient.v2_0.client.Client', mock.Mock())
    def _mock_tag_task(self, mock_task):
        mock_task.axapi_client = self.client_mock
        mock_task.get_vlan_id = mock.Mock()
        mock_task.get_vlan_id.return_value = VLAN_ID
        mock_task._get_ve_ip = mock.Mock()
        mock_task._get_ve_ip.return_value = VE_IP
        mock_task.get_subnet_and_mask = mock.Mock()
        mock_task.get_subnet_and_mask.return_value = [
            SUBNET_MASK[0], SUBNET_MASK[1]]
        mock_task._subnet_ip = SUBNET_MASK[0]
        mock_task._subnet_mask = SUBNET_MASK[1]
        mock_task._subnet = mock.Mock()
        mock_task._subnet.cidr = SUBNET_MASK[2]
        mock_task._subnet = mock.Mock()
        mock_task._subnet.id = mock.Mock()
        mock_task._subnet.network_id = "mock-network-1"
        mock_task._network_driver = a10_octavia_neutron.A10OctaviaNeutronDriver()

    def test_TagInterfaceForLB_create_vlan_ve_with_dhcp(self):
        intf = a10_utils.convert_interface_to_data_model(ETHERNET_INTERFACE)
        device1_network_map = a10_data_models.DeviceNetworkMap(vcs_device_id=1,
                                                               ethernet_interfaces=[intf])
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.device_network_map = [device1_network_map]
        lb = self._mock_lb()
        mock_task = task.TagInterfaceForLB()
        self._mock_tag_task(mock_task)
        self.client_mock.interface.ethernet.get.return_value = ETH_DATA
        self.client_mock.vlan.exists.return_value = False
        mock_task._network_driver.neutron_client.create_port = mock.Mock()
        utils.convert_port_dict_to_model = mock.Mock()
        mock_task._network_driver.get_network = mock.Mock()
        mock_task._network_driver.get_network.return_value = NETWORK_11
        mock_task._network_driver.list_networks = mock.Mock()
        mock_task._network_driver.list_networks.return_value = [NETWORK_11]
        mock_task.execute(lb, mock_thunder)
        self.client_mock.vlan.create.assert_called_with(VLAN_ID,
                                                        tagged_eths=[
                                                            TAG_INTERFACE],
                                                        veth=True)
        args, kwargs = mock_task._network_driver.neutron_client.create_port.call_args
        self.assertIn('port', args[0])
        port = args[0]['port']
        fixed_ip = port['fixed_ips'][0]
        self.assertEqual(VE_IP, fixed_ip['ip_address'])

    def test_TagInterfaceForLB_create_ethernet_vlan_ve_with_static_ip(self):
        intf = a10_utils.convert_interface_to_data_model(ETHERNET_INTERFACE)
        device1_network_map = a10_data_models.DeviceNetworkMap(vcs_device_id=1,
                                                               ethernet_interfaces=[intf])
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.device_network_map = [device1_network_map]
        lb = self._mock_lb()
        mock_task = task.TagInterfaceForLB()
        self._mock_tag_task(mock_task)
        mock_task.get_vlan_id.return_value = VE_IP_VLAN_ID
        mock_task._subnet_ip = VE_IP_SUBNET_MASK[0]
        mock_task._subnet_mask = VE_IP_SUBNET_MASK[1]
        mock_task._subnet.cidr = VE_IP_SUBNET_MASK[2]
        self.client_mock.interface.ethernet.get.return_value = ETH_DATA
        self.client_mock.vlan.exists.return_value = False
        mock_task._network_driver.neutron_client.create_port = mock.Mock()
        mock_task._network_driver.get_network = mock.Mock()
        mock_task._network_driver.get_network.return_value = NETWORK_12
        mock_task._network_driver.list_networks = mock.Mock()
        mock_task._network_driver.list_networks.return_value = [NETWORK_12]
        mock_task.execute(lb, mock_thunder)
        self.client_mock.vlan.create.assert_called_with(VE_IP_VLAN_ID,
                                                        tagged_eths=[
                                                            TAG_INTERFACE],
                                                        veth=True)
        self.client_mock.interface.ve.create.assert_called_with(VE_IP_VLAN_ID,
                                                                ip_address=STATIC_VE_IP,
                                                                ip_netmask="255.255.255.0",
                                                                enable=True)

    def test_TagInterfaceForLB_create_trunk_vlan_ve_with_static_ip(self):
        intf = a10_utils.convert_interface_to_data_model(TRUNK_INTERFACE)
        device1_network_map = a10_data_models.DeviceNetworkMap(vcs_device_id=1,
                                                               trunk_interfaces=[intf])
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.device_network_map = [device1_network_map]
        lb = self._mock_lb()
        mock_task = task.TagInterfaceForLB()
        self._mock_tag_task(mock_task)
        mock_task.get_vlan_id.return_value = VE_IP_VLAN_ID
        mock_task._subnet_ip = VE_IP_SUBNET_MASK[0]
        mock_task._subnet_mask = VE_IP_SUBNET_MASK[1]
        mock_task._subnet.cidr = VE_IP_SUBNET_MASK[2]
        self.client_mock.interface.ethernet.get.return_value = ETH_DATA
        self.client_mock.vlan.exists.return_value = False
        mock_task._network_driver.neutron_client.create_port = mock.Mock()
        mock_task._network_driver.get_network = mock.Mock()
        mock_task._network_driver.get_network.return_value = NETWORK_12
        mock_task._network_driver.list_networks = mock.Mock()
        mock_task._network_driver.list_networks.return_value = [NETWORK_12]
        mock_task.execute(lb, mock_thunder)
        self.client_mock.vlan.create.assert_called_with(VE_IP_VLAN_ID,
                                                        tagged_trunks=[
                                                            TAG_TRUNK_INTERFACE],
                                                        veth=True)
        self.client_mock.interface.ve.create.assert_called_with(VE_IP_VLAN_ID,
                                                                ip_address=STATIC_VE_IP,
                                                                ip_netmask="255.255.255.0",
                                                                enable=True)

    def test_TagInterfaceForLB_create_vlan_ve_with_axapi_exception(self):
        intf = a10_utils.convert_interface_to_data_model(ETHERNET_INTERFACE)
        device1_network_map = a10_data_models.DeviceNetworkMap(vcs_device_id=1,
                                                               ethernet_interfaces=[intf])
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.device_network_map = [device1_network_map]
        lb = self._mock_lb()
        mock_task = task.TagInterfaceForLB()
        self._mock_tag_task(mock_task)
        self.client_mock.interface.ethernet.get.return_value = ETH_DATA
        self.client_mock.vlan.exists.side_effect = Exception
        self.assertRaises(
            Exception, lambda: mock_task.execute(lb, mock_thunder))

    def test_TagInterfaceForLB_revert_created_vlan(self):
        intf = a10_utils.convert_interface_to_data_model(ETHERNET_INTERFACE)
        device1_network_map = a10_data_models.DeviceNetworkMap(vcs_device_id=1,
                                                               ethernet_interfaces=[intf])
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.device_network_map = [device1_network_map]
        lb = self._mock_lb()
        mock_task = task.TagInterfaceForLB()
        self._mock_tag_task(mock_task)
        mock_task.axapi_client.slb.virtual_server.get.return_value = DEL_VS_LIST
        mock_task.axapi_client.slb.server.get.return_value = DEL_SERVER_LIST
        mock_task._network_driver.get_port_id_from_ip = mock.Mock()
        mock_task._network_driver.neutron_client.delete_port = mock.Mock()
        mock_task._network_driver.get_port_id_from_ip.return_value = DEL_PORT_ID
        mock_task.revert(lb, mock_thunder)
        self.client_mock.vlan.delete.assert_called_with(VLAN_ID)
        mock_task._network_driver.neutron_client.delete_port.assert_called_with(
            DEL_PORT_ID)

    def test_TagInterfaceForMember_create_vlan_ve_with_dhcp(self):
        intf = a10_utils.convert_interface_to_data_model(ETHERNET_INTERFACE)
        device1_network_map = a10_data_models.DeviceNetworkMap(vcs_device_id=1,
                                                               ethernet_interfaces=[intf])
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.device_network_map = [device1_network_map]
        member = self._mock_member()
        mock_task = task.TagInterfaceForMember()
        self._mock_tag_task(mock_task)
        self.client_mock.interface.ethernet.get.return_value = ETH_DATA
        self.client_mock.vlan.exists.return_value = False
        mock_task._network_driver.neutron_client.create_port = mock.Mock()
        mock_task._network_driver.get_network = mock.Mock()
        mock_task._network_driver.get_network.return_value = NETWORK_11
        mock_task._network_driver.list_networks = mock.Mock()
        mock_task._network_driver.list_networks.return_value = [NETWORK_11]
        mock_task.execute(member, mock_thunder)
        self.client_mock.vlan.create.assert_called_with(VLAN_ID,
                                                        tagged_eths=[
                                                            TAG_INTERFACE],
                                                        veth=True)
        args, kwargs = mock_task._network_driver.neutron_client.create_port.call_args
        self.assertIn('port', args[0])
        port = args[0]['port']
        fixed_ip = port['fixed_ips'][0]
        self.assertEqual(VE_IP, fixed_ip['ip_address'])

    def test_TagInterfaceForMember_create_vlan_ve_with_neutron_exception(self):
        intf = a10_utils.convert_interface_to_data_model(ETHERNET_INTERFACE)
        device1_network_map = a10_data_models.DeviceNetworkMap(vcs_device_id=1,
                                                               ethernet_interfaces=[intf])
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.device_network_map = [device1_network_map]
        member = self._mock_member()
        mock_task = task.TagInterfaceForMember()
        self._mock_tag_task(mock_task)
        self.client_mock.interface.ethernet.get.return_value = ETH_DATA
        self.client_mock.vlan.exists.return_value = False
        mock_task._network_driver.neutron_client.create_port = mock.Mock()
        mock_task._network_driver.neutron_client.create_port.side_effect = Exception
        self.assertRaises(
            Exception, lambda: mock_task.execute(member, mock_thunder))

    def test_TagInterfaceForMember_revert_created_vlan(self):
        intf = a10_utils.convert_interface_to_data_model(ETHERNET_INTERFACE)
        device1_network_map = a10_data_models.DeviceNetworkMap(vcs_device_id=1,
                                                               ethernet_interfaces=[intf])
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.device_network_map = [device1_network_map]
        member = self._mock_member()
        mock_task = task.TagInterfaceForMember()
        self._mock_tag_task(mock_task)
        mock_task.axapi_client.slb.virtual_server.get.return_value = DEL_VS_LIST
        mock_task.axapi_client.slb.server.get.return_value = DEL_SERVER_LIST
        mock_task._network_driver.get_port_id_from_ip = mock.Mock()
        mock_task._network_driver.neutron_client.delete_port = mock.Mock()
        mock_task._network_driver.get_port_id_from_ip.return_value = DEL_PORT_ID
        mock_task.revert(member, mock_thunder)
        self.client_mock.vlan.delete.assert_called_with(VLAN_ID)
        mock_task._network_driver.neutron_client.delete_port.assert_called_with(
            DEL_PORT_ID)

    def test_TagInterfaceForMember_execute_log_warning_no_subnet_id(self):
        member = self._mock_member()
        member.id = "1234"
        member.subnet_id = None
        task_path = "a10_octavia.controller.worker.tasks.vthunder_tasks"
        log_message = str("Subnet id argument was not specified during "
                          "issuance of create command/API call for member %s. "
                          "Skipping TagInterfaceForMember task")
        expected_log = ["WARNING:{}:{}".format(
            task_path, log_message % member.id)]
        mock_task = task.TagInterfaceForMember()
        with self.assertLogs(task_path, level='WARN') as cm:
            mock_task.execute(member, VTHUNDER)
            self.assertEqual(expected_log, cm.output)

    def test_TagInterfaceForMember_revert_log_warning_no_subnet_id(self):
        member = self._mock_member()
        member.id = "1234"
        member.subnet_id = None
        task_path = "a10_octavia.controller.worker.tasks.vthunder_tasks"
        log_message = str("Subnet id argument was not specified during "
                          "issuance of create command/API call for member %s. "
                          "Skipping TagInterfaceForMember task")
        expected_log = ["WARNING:{}:{}".format(
            task_path, log_message % member.id)]
        mock_task = task.TagInterfaceForMember()
        with self.assertLogs(task_path, level='WARN') as cm:
            mock_task.revert(member, VTHUNDER)
            self.assertEqual(expected_log, cm.output)

    def test_DeleteInterfaceTagIfNotInUseForMember_execute_log_warning_no_subnet_id(self):
        member = self._mock_member()
        member.id = "1234"
        member.subnet_id = None
        task_path = "a10_octavia.controller.worker.tasks.vthunder_tasks"
        log_message = str("Subnet id argument was not specified during "
                          "issuance of create command/API call for member %s. "
                          "Skipping DeleteInterfaceTagIfNotInUseForMember task")
        expected_log = ["WARNING:{}:{}".format(
            task_path, log_message % member.id)]
        mock_task = task.DeleteInterfaceTagIfNotInUseForMember()
        with self.assertLogs(task_path, level='WARN') as cm:
            mock_task.execute(member, VTHUNDER)
            self.assertEqual(expected_log, cm.output)

    def test_DeleteInterfaceTagIfNotInUseForLB_execute_delete_vlan(self):
        intf = a10_utils.convert_interface_to_data_model(ETHERNET_INTERFACE)
        device1_network_map = a10_data_models.DeviceNetworkMap(vcs_device_id=1,
                                                               ethernet_interfaces=[intf])
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.device_network_map = [device1_network_map]
        lb = self._mock_lb()
        mock_task = task.DeleteInterfaceTagIfNotInUseForLB()
        self._mock_tag_task(mock_task)
        mock_task.axapi_client.slb.virtual_server.get.return_value = DEL_VS_LIST
        mock_task.axapi_client.slb.server.get.return_value = DEL_SERVER_LIST
        mock_task._network_driver.get_port_id_from_ip = mock.Mock()
        mock_task._network_driver.neutron_client.delete_port = mock.Mock()
        mock_task._network_driver.get_port_id_from_ip.return_value = DEL_PORT_ID
        mock_task.execute(lb, mock_thunder)
        self.client_mock.vlan.delete.assert_called_with(VLAN_ID)
        mock_task._network_driver.neutron_client.delete_port.assert_called_with(
            DEL_PORT_ID)

    def test_DeleteInterfaceTagIfNotInUseForMember_execute_delete_vlan(self):
        intf = a10_utils.convert_interface_to_data_model(ETHERNET_INTERFACE)
        device1_network_map = a10_data_models.DeviceNetworkMap(vcs_device_id=1,
                                                               ethernet_interfaces=[intf])
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.device_network_map = [device1_network_map]
        member = self._mock_member()
        mock_task = task.DeleteInterfaceTagIfNotInUseForMember()
        self._mock_tag_task(mock_task)
        mock_task.axapi_client.slb.virtual_server.get.return_value = DEL_VS_LIST
        mock_task.axapi_client.slb.server.get.return_value = DEL_SERVER_LIST
        mock_task._network_driver.get_port_id_from_ip = mock.Mock()
        mock_task._network_driver.neutron_client.delete_port = mock.Mock()
        mock_task._network_driver.get_port_id_from_ip.return_value = DEL_PORT_ID
        mock_task.execute(member, mock_thunder)
        self.client_mock.vlan.delete.assert_called_with(VLAN_ID)
        mock_task._network_driver.neutron_client.delete_port.assert_called_with(
            DEL_PORT_ID)

    def test_SetupDeviceNetworkMap_execute_vcs_disabled(self):
        self.conf.register_opts(config_options.A10_HARDWARE_THUNDER_OPTS,
                                group=a10constants.A10_HARDWARE_THUNDER_CONF_SECTION)
        devices_str = json.dumps([RACK_DEVICE])
        self.conf.config(group=a10constants.A10_HARDWARE_THUNDER_CONF_SECTION,
                         devices=devices_str)
        mock_task = task.SetupDeviceNetworkMap()
        self._mock_tag_task(mock_task)
        mock_task.axapi_client.system.action.get_vcs_summary_oper.return_value = VCS_DISABLED
        mock_thunder = copy.deepcopy(VTHUNDER)
        thunder = mock_task.execute(mock_thunder)
        self.assertEqual(len(thunder.device_network_map), 1)
        self.assertEqual(thunder.device_network_map[0].vcs_device_id, 1)

    def test_SetupDeviceNetworkMap_execute_vcs_master_vblade(self):
        self.conf.register_opts(config_options.A10_HARDWARE_THUNDER_OPTS,
                                group=a10constants.A10_HARDWARE_THUNDER_CONF_SECTION)

        devices_str = json.dumps([RACK_DEVICE_VCS])
        self.conf.config(group=a10constants.A10_HARDWARE_THUNDER_CONF_SECTION,
                         devices=devices_str)
        mock_task = task.SetupDeviceNetworkMap()
        self._mock_tag_task(mock_task)
        mock_task.axapi_client.system.action.get_vcs_summary_oper.return_value = VCS_MASTER_VBLADE
        mock_thunder = copy.deepcopy(VTHUNDER)
        thunder = mock_task.execute(mock_thunder)
        self.assertEqual(len(thunder.device_network_map), 2)
        self.assertEqual(thunder.device_network_map[0].vcs_device_id, 1)
        self.assertEqual(thunder.device_network_map[0].state, "Master")
        self.assertEqual(
            thunder.device_network_map[0].mgmt_ip_address, DEVICE1_MGMT_IP)
        self.assertEqual(thunder.device_network_map[1].vcs_device_id, 2)
        self.assertEqual(thunder.device_network_map[1].state, "vBlade")
        self.assertEqual(
            thunder.device_network_map[1].mgmt_ip_address, DEVICE2_MGMT_IP)

    def test_SetupDeviceNetworkMap_execute_vcs_device1_failed(self):
        self.conf.register_opts(config_options.A10_HARDWARE_THUNDER_OPTS,
                                group=a10constants.A10_HARDWARE_THUNDER_CONF_SECTION)
        devices_str = json.dumps([RACK_DEVICE_VCS])
        self.conf.config(group=a10constants.A10_HARDWARE_THUNDER_CONF_SECTION,
                         devices=devices_str)
        mock_task = task.SetupDeviceNetworkMap()
        self._mock_tag_task(mock_task)
        mock_task.axapi_client.system.action.get_vcs_summary_oper.return_value = VCS_DEVICE1_FAILED
        mock_thunder = copy.deepcopy(VTHUNDER)
        thunder = mock_task.execute(mock_thunder)
        self.assertEqual(len(thunder.device_network_map), 1)
        self.assertEqual(thunder.device_network_map[0].vcs_device_id, 2)
        self.assertEqual(thunder.device_network_map[0].state, "Master")
        self.assertEqual(
            thunder.device_network_map[0].mgmt_ip_address, DEVICE2_MGMT_IP)

    def test_SetupDeviceNetworkMap_execute_delete_flow_after_error_no_fail(self):
        ret_val = task.SetupDeviceNetworkMap().execute(vthunder=None)
        self.assertIsNone(ret_val)

    def test_WriteMemory_execute_save_shared_mem(self):
        self.conf.register_opts(config_options.A10_HOUSE_KEEPING_OPTS,
                                group=a10constants.A10_HOUSE_KEEPING)
        self.conf.config(group=a10constants.A10_HOUSE_KEEPING,
                         use_periodic_write_memory='disable')
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_task = task.WriteMemory()
        mock_task.axapi_client = self.client_mock
        mock_task.execute(mock_thunder)
        self.client_mock.system.action.write_memory.assert_called_with(
            partition='shared')

    def test_WriteMemory_execute_save_specific_partition_mem(self):
        self.conf.register_opts(config_options.A10_HOUSE_KEEPING_OPTS,
                                group=a10constants.A10_HOUSE_KEEPING)
        self.conf.config(group=a10constants.A10_HOUSE_KEEPING,
                         use_periodic_write_memory='disable')
        thunder = copy.deepcopy(VTHUNDER)
        thunder.partition_name = "testPartition"
        mock_task = task.WriteMemory()
        mock_task.axapi_client = self.client_mock
        mock_task.execute(thunder)
        self.client_mock.system.action.write_memory.assert_called_with(
            partition='specified',
            specified_partition='testPartition')

    def test_WriteMemoryHouseKeeper_execute_save_shared_mem(self):
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.loadbalancer_id = a10constants.MOCK_LOAD_BALANCER_ID
        lb_list = []
        lb_list.append(LB)
        mock_task = task.WriteMemoryHouseKeeper()
        mock_task.axapi_client = self.client_mock
        mock_task.execute(mock_thunder, lb_list, True)
        self.client_mock.system.action.write_memory.assert_called_with(
            partition='shared')

    def test_WriteMemoryHouseKeeper_execute_save_specific_partition_mem(self):
        thunder = copy.deepcopy(VTHUNDER)
        thunder.loadbalancer_id = a10constants.MOCK_LOAD_BALANCER_ID
        lb_list = []
        lb_list.append(LB)
        thunder.partition_name = "testPartition"
        mock_task = task.WriteMemoryHouseKeeper()
        mock_task.axapi_client = self.client_mock
        mock_task.execute(thunder, lb_list)
        self.client_mock.system.action.write_memory.assert_called_with(
            partition='specified',
            specified_partition='testPartition')

    def test_WriteMemory_execute_not_called(self):
        self.conf.register_opts(config_options.A10_HOUSE_KEEPING_OPTS,
                                group=a10constants.A10_HOUSE_KEEPING)
        self.conf.config(group=a10constants.A10_HOUSE_KEEPING,
                         use_periodic_write_memory='enable')
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_task = task.WriteMemory()
        mock_task.axapi_client = self.client_mock
        mock_task.execute(mock_thunder)
        self.client_mock.system.action.write_memory.assert_not_called()

    def test_WriteMemory_execute_delete_flow_after_error_no_fail(self):
        self.conf.register_opts(config_options.A10_HOUSE_KEEPING_OPTS,
                                group=a10constants.A10_HOUSE_KEEPING)
        self.conf.config(group=a10constants.A10_HOUSE_KEEPING,
                         use_periodic_write_memory='enable')
        ret_val = task.WriteMemory().execute(vthunder=None)
        self.assertIsNone(ret_val)

    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.a10_utils')
    def test_HandleACOSPartitionChange_execute_create_partition(self, mock_utils):
        mock_client = mock.Mock()
        mock_client.system.partition.get.side_effect = acos_errors.NotFound()
        mock_utils.get_axapi_client.return_value = mock_client
        mock_thunder_config = copy.deepcopy(VTHUNDER)
        mock_thunder_config.partition_name = "PartitionA"
        task.HandleACOSPartitionChange().execute(LB, mock_thunder_config)
        mock_client.system.partition.create.assert_called_with("PartitionA")

    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.a10_utils')
    def test_HandleACOSPartitionChange_execute_create_partition_fail(self, mock_utils):
        mock_client = mock.Mock()
        mock_client.system.partition.get.side_effect = acos_errors.NotFound()
        mock_client.system.action.write_memory.side_effect = acos_errors.ACOSException()
        mock_utils.get_axapi_client.return_value = mock_client
        mock_thunder_config = copy.deepcopy(VTHUNDER)
        mock_thunder_config.partition_name = "PartitionA"
        expected_error = acos_errors.ACOSException
        task_function = task.HandleACOSPartitionChange().execute
        self.assertRaises(expected_error, task_function, LB, mock_thunder_config)

    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.a10_utils')
    def test_HandleACOSPartitionChange_execute_partition_found_active(self, mock_utils):
        partition = {
            'partition-name': 'PartitionA',
            'status': 'Active'
        }
        mock_client = mock.Mock()
        mock_client.system.partition.get.return_value = partition
        mock_utils.get_axapi_client.return_value = mock_client
        mock_thunder_config = copy.deepcopy(VTHUNDER)
        mock_thunder_config.partition_name = "PartitionA"
        task.HandleACOSPartitionChange().execute(LB, mock_thunder_config)
        mock_client.system.partition.create.assert_not_called()

    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.a10_utils')
    def test_HandleACOSPartitionChange_execute_partition_found_not_active(self, mock_utils):
        partition = {
            'partition-name': 'PartitionA',
            'status': 'Not-Active'
        }
        mock_client = mock.Mock()
        mock_client.system.partition.get.return_value = partition
        mock_utils.get_axapi_client.return_value = mock_client
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_thunder.partition_name = "PartitionA"
        expected_error = exceptions.PartitionNotActiveError
        task_function = task.HandleACOSPartitionChange().execute
        self.assertRaises(expected_error, task_function, LB, mock_thunder)

    @mock.patch('a10_octavia.common.utils.get_parent_project',
                return_value=None)
    def test_HandleACOSPartitionChange_parent_partition_not_exists(
            self, mock_parent_project_id):
        self.conf.config(
            group=a10constants.A10_GLOBAL_CONF_SECTION,
            use_parent_partition=True)
        mock_lb = copy.deepcopy(LB)
        mock_lb.project_id = a10constants.MOCK_CHILD_PROJECT_ID
        mock_thunder_config = copy.deepcopy(VTHUNDER)
        mock_thunder_config.hierarchical_multitenancy = "enable"
        expected_error = exceptions.ParentProjectNotFound
        task_function = task.HandleACOSPartitionChange().execute
        self.assertRaises(expected_error, task_function, mock_lb, mock_thunder_config)

    @mock.patch('a10_octavia.common.utils.get_parent_project',
                return_value=a10constants.MOCK_PARENT_PROJECT_ID)
    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.a10_utils')
    def test_HandleACOSPartitionChange_parent_partition_exists(
            self, mock_utils, mock_parent_project_id):
        self.conf.config(
            group=a10constants.A10_GLOBAL_CONF_SECTION,
            use_parent_partition=True)

        partition = {
            'partition-name': a10constants.MOCK_PARENT_PROJECT_ID[:14],
            'status': 'Active'
        }
        mock_client = mock.Mock()
        mock_client.system.partition.get.return_value = partition
        mock_utils.get_axapi_client.return_value = mock_client

        mock_lb = copy.deepcopy(LB)
        mock_lb.project_id = a10constants.MOCK_CHILD_PROJECT_ID
        mock_thunder_config = copy.deepcopy(VTHUNDER)
        mock_thunder_config.hierarchical_multitenancy = "enable"
        vthunder_config = task.HandleACOSPartitionChange().execute(
            LB, mock_thunder_config)
        self.assertEqual(vthunder_config.partition_name,
                         a10constants.MOCK_PARENT_PROJECT_ID[:14])

    @mock.patch('a10_octavia.common.utils.get_parent_project',
                return_value=a10constants.MOCK_PARENT_PROJECT_ID)
    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.a10_utils')
    def test_HandleACOSPartitionChange_parent_partition_no_hmt(
            self, mock_utils, mock_parent_project_id):
        self.conf.config(group=a10constants.A10_GLOBAL_CONF_SECTION,
                         use_parent_partition=True)
        partition = {
            'partition-name': a10constants.MOCK_CHILD_PROJECT_ID[:14],
            'status': 'Active'
        }
        mock_client = mock.Mock()
        mock_client.system.partition.get.return_value = partition
        mock_utils.get_axapi_client.return_value = mock_client

        mock_lb = copy.deepcopy(LB)
        mock_lb.project_id = a10constants.MOCK_CHILD_PROJECT_ID
        mock_thunder_config = copy.deepcopy(VTHUNDER)
        mock_thunder_config.hierarchical_multitenancy = "disable"
        vthunder_config = task.HandleACOSPartitionChange().execute(
            LB, mock_thunder_config)

        self.assertEqual(vthunder_config.partition_name, a10constants.MOCK_CHILD_PROJECT_ID[:14])

    @mock.patch('a10_octavia.common.utils.get_parent_project',
                return_value=a10constants.MOCK_PARENT_PROJECT_ID)
    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.a10_utils')
    def test_HandleACOSPartitionChange_parent_partition_hmt_no_use_parent_partition(
            self, mock_utils, mock_parent_project_id):
        self.conf.config(
            group=a10constants.A10_GLOBAL_CONF_SECTION,
            use_parent_partition=False)

        partition = {
            'partition-name': a10constants.MOCK_CHILD_PROJECT_ID[:14],
            'status': 'Active'
        }
        mock_client = mock.Mock()
        mock_client.system.partition.get.return_value = partition
        mock_utils.get_axapi_client.return_value = mock_client

        mock_lb = copy.deepcopy(LB)
        mock_lb.project_id = a10constants.MOCK_CHILD_PROJECT_ID
        mock_thunder_config = copy.deepcopy(VTHUNDER)
        mock_thunder_config.hierarchical_multitenancy = "enable"

        vthunder_config = task.HandleACOSPartitionChange().execute(
            LB, mock_thunder_config)
        self.assertEqual(vthunder_config.partition_name, a10constants.MOCK_CHILD_PROJECT_ID[:14])

    @mock.patch('a10_octavia.common.utils.get_parent_project',
                return_value=a10constants.MOCK_PARENT_PROJECT_ID)
    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.a10_utils')
    def test_HandleACOSPartitionChange_child_partition_exists(
            self, mock_utils, mock_parent_project_id):
        self.conf.config(
            group=a10constants.A10_GLOBAL_CONF_SECTION,
            use_parent_partition=False)

        partition = {
            'partition-name': a10constants.MOCK_CHILD_PROJECT_ID[:14],
            'status': 'Active'
        }
        mock_client = mock.Mock()
        mock_client.system.partition.get.return_value = partition
        mock_utils.get_axapi_client.return_value = mock_client

        mock_lb = copy.deepcopy(LB)
        mock_lb.project_id = a10constants.MOCK_CHILD_PROJECT_ID
        mock_thunder_config = copy.deepcopy(VTHUNDER)
        mock_thunder_config.hierarchical_multitenancy = "enable"

        vthunder_config = task.HandleACOSPartitionChange().execute(
            LB, mock_thunder_config)
        self.assertEqual(vthunder_config.partition_name,
                         a10constants.MOCK_CHILD_PROJECT_ID[:14])

    @mock.patch('a10_octavia.common.utils.get_parent_project',
                return_value=a10constants.MOCK_PARENT_PROJECT_ID)
    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.a10_utils')
    def test_HandleACOSPartitionChange_nlbaas_backwards_compat(
            self, mock_utils, mock_parent_project_id):
        self.conf.config(
            group=a10constants.A10_GLOBAL_CONF_SECTION,
            use_parent_partition=False)

        partition = {
            'partition-name': a10constants.MOCK_CHILD_PROJECT_ID[:13],
            'status': 'Active'
        }
        mock_client = mock.Mock()
        mock_client.system.partition.get.return_value = partition
        mock_utils.get_axapi_client.return_value = mock_client

        mock_lb = copy.deepcopy(LB)
        mock_lb.project_id = a10constants.MOCK_CHILD_PROJECT_ID
        mock_thunder_config = copy.deepcopy(VTHUNDER)
        mock_thunder_config.hierarchical_multitenancy = "enable"

        vthunder_config = task.HandleACOSPartitionChange().execute(
            LB, mock_thunder_config)
        self.assertEqual(vthunder_config.partition_name,
                         a10constants.MOCK_CHILD_PROJECT_ID[:13])

    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.time')
    def test_AmphoraPostVipPlug_execute_for_reload_reboot(self, mock_time):
        thunder = copy.deepcopy(VTHUNDER)
        added_ports = {'amphora_id': '123'}
        mock_task = task.AmphoraePostVIPPlug()
        mock_task.axapi_client = self.client_mock
        mock_task.loadbalancer_repo = mock.MagicMock()
        mock_task.vthunder_repo = mock.MagicMock()
        mock_task.loadbalancer_repo.check_lb_with_distinct_subnet_and_project.return_value = True
        mock_task.vthunder_repo.get_vthunder_by_project_id_and_role.return_value = thunder
        vthunder = mock_task.vthunder_repo.get_vthunder_by_project_id_and_role.return_value
        mock_task.execute(LB, thunder, added_ports)
        self.client_mock.system.action.write_memory.assert_called_with()
        self.client_mock.system.action.reload_reboot_for_interface_attachment.assert_called_with(
            vthunder.acos_version)

    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.time')
    def test_AmphoraPostVipPlug_execute_for_no_reload_reboot(self, mock_time):
        thunder = copy.deepcopy(VTHUNDER)
        added_ports = {'amphora_id': ''}
        mock_task = task.AmphoraePostVIPPlug()
        mock_task.axapi_client = self.client_mock
        mock_task.loadbalancer_repo = mock.MagicMock()
        mock_task.loadbalancer_repo.check_lb_with_distinct_subnet_and_project.return_value = False
        mock_task.execute(LB, thunder, added_ports)
        self.client_mock.system.action.write_memory.assert_not_called()
        self.client_mock.system.action.reload_reboot_for_interface_attachment.assert_not_called()

    def test_UpdateAcosVersionInVthunderEntry_execute_for_first_lb(self):
        thunder = copy.deepcopy(VTHUNDER)
        mock_task = task.UpdateAcosVersionInVthunderEntry()
        mock_task.axapi_client = mock.MagicMock()[1:999]
        mock_task.vthunder_repo = mock.MagicMock()[1:999]
        mock_task.vthunder_repo.get_vthunder_by_project_id.return_value = None
        mock_task.vthunder_repo.get_delete_compute_flag.return_value = True
        mock_task.execute(thunder, LB)
        mock_task.axapi_client.system.action.get_acos_version.assert_called_with()

    def test_UpdateAcosVersionInVthunderEntry_execute_for_lb_with_existing_vthunder(self):
        thunder = copy.deepcopy(VTHUNDER)
        mock_task = task.UpdateAcosVersionInVthunderEntry()
        mock_task.axapi_client = mock.MagicMock()[1:999]
        mock_task.vthunder_repo = mock.MagicMock()[1:999]
        mock_task.vthunder_repo.get_vthunder_by_project_id.return_value = VTHUNDER
        mock_task.vthunder_repo.get_delete_compute_flag.return_value = False
        mock_task.execute(thunder, LB)
        mock_task.axapi_client.system.action.get_acos_version.assert_not_called()

    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.time')
    def test_AmphoraePostMemberNetworkPlug_execute_for_reload_reboot(self, mock_time):
        thunder = copy.deepcopy(VTHUNDER)
        thunder.acos_version = "5.2.1"
        added_ports = {'amphora_id': '123'}
        mock_task = task.AmphoraePostMemberNetworkPlug()
        mock_task.axapi_client = self.client_mock
        mock_task.execute(added_ports, LB, thunder)
        self.client_mock.system.action.write_memory.assert_called_with()
        self.client_mock.system.action.reload_reboot_for_interface_attachment.assert_called_with(
            "5.2.1")

    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.time')
    def test_AmphoraePostMemberNetworkPlug_execute_for_no_reload_reboot(self, mock_time):
        thunder = copy.deepcopy(VTHUNDER)
        thunder.acos_version = "5.2.1"
        added_ports = {'amphora_id': ''}
        mock_task = task.AmphoraePostMemberNetworkPlug()
        mock_task.axapi_client = self.client_mock
        mock_task.execute(added_ports, LB, thunder)
        self.client_mock.system.action.write_memory.assert_not_called()
        self.client_mock.system.action.reload_reboot_for_interface_attachment.assert_not_called()

    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.time')
    def test_AmphoraePostNetworkUnplug_execute_for_reload_reboot(self, mock_time):
        thunder = copy.deepcopy(VTHUNDER)
        added_ports = {'amphora_id': '123'}
        mock_task = task.AmphoraePostNetworkUnplug()
        mock_task.axapi_client = self.client_mock
        mock_task.execute(added_ports, LB, thunder)
        self.client_mock.system.action.write_memory.assert_called_with()
        self.client_mock.system.action.reload_reboot_for_interface_detachment.assert_called_with(
            "5.2.1")

    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.time')
    def test_AmphoraePostNetworkUnplug_execute_no_port_no_verion_no_reload_reboot(self, mock_time):
        thunder = copy.deepcopy(VTHUNDER)
        added_ports = {'amphora_id': ''}
        mock_task = task.AmphoraePostNetworkUnplug()
        mock_task.axapi_client = self.client_mock
        mock_task.execute(added_ports, LB, thunder)
        self.client_mock.system.action.write_memory.assert_not_called()
        self.client_mock.system.action.reload_reboot_for_interface_detachment.assert_not_called()

    @mock.patch('stevedore.driver.DriverManager')
    def test_EnableInterface_execute_for_single_topology(self, mock_drivermgr):
        thunder = copy.deepcopy(VTHUNDER)
        added_ports = {'amphora_id': ''}
        thunder.topology = "SINGLE"
        mock_task = task.EnableInterface()
        mock_task.loadbalancer_repo = mock.MagicMock()
        mock_task.loadbalancer_repo.check_lb_exists_in_project.return_value = True
        mock_task.axapi_client = self.client_mock
        self.client_mock.interface.get_list.return_value = INTERFACES
        mock_driver = mock.MagicMock()
        mock_drivermgr.return_value = mock_driver
        mock_drivermgr.get_plugged_networks.return_value = [NETWORK_11]
        mock_task.execute(thunder, LB, added_ports)
        self.client_mock.system.action.setInterface.assert_not_called()

    @mock.patch('stevedore.driver.DriverManager')
    def test_EnableInterface_execute_for_active_standby_topology(self, mock_drivermgr):
        thunder = copy.deepcopy(VTHUNDER)
        added_ports = {'amphora_id': ''}
        thunder.topology = "ACTIVE_STANDBY"
        mock_task = task.EnableInterface()
        mock_task.loadbalancer_repo = mock.MagicMock()
        mock_task.loadbalancer_repo.check_lb_exists_in_project.return_value = True
        mock_task.axapi_client = self.client_mock
        self.client_mock.interface.get_list.return_value = INTERFACES
        mock_driver = mock.MagicMock()
        mock_drivermgr.return_value = mock_driver
        mock_drivermgr.get_plugged_networks.return_value = [NETWORK_11]
        mock_task.execute(thunder, LB, added_ports)
        self.client_mock.system.action.setInterface.assert_not_called()

    def test_AmphoraePostNetworkUnplug_amophora_not_available(self):
        mock_thunder = copy.deepcopy(VTHUNDER)
        mock_LB = copy.deepcopy(LB)
        mock_LB.amphorae = []
        added_ports = {}
        mock_task = task.AmphoraePostNetworkUnplug()
        mock_task.axapi_client = self.client_mock
        mock_task.execute(added_ports, mock_LB, mock_thunder)
        self.client_mock.system.action.write_memory.assert_not_called()
        self.client_mock.system.action.reload_reboot_for_interface_detachment.assert_not_called()

    def test_GetVthunderConfByFlavor_with_flavor(self):
        conf = {}
        use_dev_flavor = False
        mock_LB = copy.deepcopy(LB)
        mock_task = task.GetVthunderConfByFlavor()
        mock_task.axapi_client = self.client_mock
        conf, use_dev_flavor = mock_task.execute(mock_LB, FLAVOR_DEV2, DEV_CONF_DICT, DEV_FLAVOR)
        self.assertEqual(use_dev_flavor, True)
        self.assertEqual(conf, FLAVOR_DEV1)

    def test_GetVthunderConfByFlavor_without_flavor(self):
        conf = {}
        use_dev_flavor = False
        mock_LB = copy.deepcopy(LB)
        mock_task = task.GetVthunderConfByFlavor()
        mock_task.axapi_client = self.client_mock
        conf, use_dev_flavor = mock_task.execute(mock_LB, FLAVOR_DEV2, DEV_CONF_DICT, None)
        self.assertEqual(use_dev_flavor, False)
        self.assertEqual(conf, FLAVOR_DEV2)

    def test_allowl2dsr_with_deployment_flavor(self):
        mock_task = task.AllowL2DSR()
        mock_task._network_driver = self.client_mock
        lb_count = 1
        mock_task.execute(SUBNET, AMPHORAE, lb_count, flavor_data=DEPLOYMENT_FLAVOR)
        self.client_mock.allow_use_any_source_ip_on_egress.assert_called_with(
            SUBNET.network_id, AMPHORAE[0])

    def test_allowl2dsr_without_deployment_flavor(self):
        mock_task = task.AllowL2DSR()
        mock_task._network_driver = self.client_mock
        lb_count = 1
        mock_task.execute(SUBNET, AMPHORAE, lb_count, flavor_data=DEV_FLAVOR)
        self.client_mock.allow_use_any_source_ip_on_egress.assert_not_called()

    def test_deletel2dsr_with_deployment_flavor(self):
        mock_task = task.DeleteL2DSR()
        mock_task._network_driver = self.client_mock
        lb_count = 1
        mock_task.execute(SUBNET, AMPHORAE, lb_count, lb_count, flavor_data=DEPLOYMENT_FLAVOR)
        self.client_mock.remove_any_source_ip_on_egress.assert_called_with(
            SUBNET.network_id, AMPHORAE[0])

    def test_get_listeners_stats(self):
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.loadbalancer_id = a10constants.MOCK_LOAD_BALANCER_ID
        lb = copy.deepcopy(LB)
        lb.provisioning_status = "ACTIVE"
        STATS = {
            "port-list": [
                {
                    "stats": {
                        "curr_conn": 0, "total_conn": 1,
                        "total_fwd_bytes": 684, "total_rev_bytes": 705
                    },
                    "port-number": 80,
                    "protocol": "http"
                }
            ]
        }
        mock_task = task.GetListenersStats()
        mock_task.loadbalancer_repo = mock.MagicMock()
        mock_task.loadbalancer_repo.get.return_value = lb
        mock_task.axapi_client = self.client_mock
        mock_task.axapi_client.slb.virtual_server.stats.return_value = STATS
        mock_task.execute(vthunder)
        self.client_mock.slb.virtual_server.stats.assert_called_with(
            name=a10constants.MOCK_LOAD_BALANCER_ID, timeout=mock.ANY, max_retries=mock.ANY)

    @mock.patch('a10_octavia.controller.worker.tasks.vthunder_tasks.time')
    def test_SparePostNetowrkPlug(self, mock_time):
        mock_task = task.SparePostNetworkPlug()
        mock_task.axapi_client = self.client_mock
        thunder = copy.deepcopy(VTHUNDER)
        net_list = [1, 2]
        mock_task.execute(thunder, net_list)
        self.client_mock.system.action.write_memory.assert_called_with()
        self.client_mock.system.action.reload_reboot_for_interface_attachment.assert_called_with(
            thunder.acos_version)

    def test_EnableInterfaceOnSpare(self):
        mock_task = task.EnableInterfaceOnSpare()
        mock_task.axapi_client = self.client_mock
        thunder = copy.deepcopy(VTHUNDER)
        net_list = [1, 2]
        iface = {'interface': {'ethernet-list': [
            {'action': "disable", 'ifnum': 2}
        ]}}
        mock_task.axapi_client.interface.get_list.return_value = iface
        mock_task.execute(thunder, net_list)
        self.client_mock.system.action.setInterface.assert_called_with(2)

    def test_ConfigureVRRPFailover(self):
        mock_task = task.ConfigureVRRPFailover()
        mock_task.axapi_client = self.client_mock
        thunder = copy.deepcopy(VTHUNDER)
        mock_task.execute(thunder, 'device1', 'set1')
        self.client_mock.system.action.configureVRRP.assert_called_with('device1', 'set1')

    def test_ConfigureaVCSFailover(self):
        mock_task = task.ConfigureaVCSFailover()
        mock_task.axapi_client = self.client_mock
        thunder = copy.deepcopy(VTHUNDER)
        mock_task.execute(thunder, 'device1')
        self.client_mock.system.action.set_vcs_device.assert_called_with('device1', 100)
        self.client_mock.system.action.vcs_enable.assert_called_with()
        self.client_mock.system.action.write_memory.assert_called_with()
        self.client_mock.system.action.vcs_reload.assert_called_with()

    def test_SetHostName_execute_set_hostname(self):
        vthunder = copy.deepcopy(VTHUNDER)
        amphora = copy.deepcopy(AMPHORA)
        amphora.id = '295990755083049101712519384016336453749'
        interfaces = {
            'interface': {
                'ethernet-list': []
            }
        }

        expected_host = 'amphora-29599075508304910171251'

        flexpool_task = task.SetVThunderHostname()
        flexpool_task.axapi_client = self.client_mock
        flexpool_task.axapi_client.interface.get_list.return_value = interfaces

        flexpool_task.execute(vthunder, amphora)
        self.client_mock.system.action.set_hostname.assert_called_with(expected_host)

    def test_DeleteHealthMonitorOnVThunder_success(self):
        mock_task = task.DeleteHealthMonitorOnVThunder()
        mock_task.axapi_client = self.client_mock
        thunder = copy.deepcopy(VTHUNDER)
        mock_lb_count_thunder_partition = 0
        mock_task.execute(mock_lb_count_thunder_partition, thunder)
        self.client_mock.slb.server.delete.assert_called_with(
            a10constants.MOCK_OCTAVIA_HEALTH_MANAGER_CONTROLLER)
        self.client_mock.slb.hm.delete.assert_called_with(
            a10constants.MOCK_OCTAVIA_HEALTH_MONITOR)

    def test_DeleteHealthMonitorOnVThunder_no_vthunder(self):
        mock_task = task.DeleteHealthMonitorOnVThunder()
        mock_lb_count_thunder_partition = 0
        delete_hm = mock_task.execute(mock_lb_count_thunder_partition, None)
        self.assertEqual(None, delete_hm)

    def test_DeleteHealthMonitorOnVThunder_active_lb(self):
        mock_task = task.DeleteHealthMonitorOnVThunder()
        mock_lb_count_thunder_partition = 2
        thunder = copy.deepcopy(VTHUNDER)
        delete_hm = mock_task.execute(mock_lb_count_thunder_partition, thunder)
        self.assertEqual(None, delete_hm)



#    Copyright 2021, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import imp

try:
    from unittest import mock
except ImportError:
    import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

from octavia.common import data_models as o_data_models
from octavia.network import data_models as n_data_models
from octavia.tests.common import constants as t_constants

from acos_client import errors as acos_errors

from a10_octavia.common import config_options
from a10_octavia.common import data_models as a10_data_models
from a10_octavia.common import exceptions as a10_ex
from a10_octavia.controller.worker.tasks import glm_tasks as task
from a10_octavia.tests.common import a10constants
from a10_octavia.tests.unit import base


VTHUNDER = a10_data_models.VThunder(id=a10constants.MOCK_VTHUNDER_ID)
AMPHORA = o_data_models.Amphora(id=t_constants.MOCK_AMP_ID1)
DNS_SUBNET = n_data_models.Subnet(id=a10constants.MOCK_SUBNET_ID)
DNS_NETWORK = n_data_models.Network(id=a10constants.MOCK_NETWORK_ID,
                                    subnets=[DNS_SUBNET.id])
PRIMARY_DNS = '1.3.3.7'
SECONDARY_DNS = '1.0.0.7'
PROXY_HOST = '10.10.10.10'
PROXY_PORT = 1111
PROXY_USERNAME = 'user'
PROXY_PASSWORD = True
PROXY_PASSWORD_VALUE = 'password'


class TestGLMTasks(base.BaseTaskTestCase):

    def setUp(self):
        super(TestGLMTasks, self).setUp()
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        self.conf.register_opts(config_options.A10_GLM_LICENSE_OPTS,
                                group=a10constants.A10_GLOBAL_CONF_SECTION)
        imp.reload(task)
        self.client_mock = mock.Mock()
        self.db_session = mock.patch(
            'a10_octavia.controller.worker.tasks.a10_database_tasks.db_apis.get_session')
        self.db_session.start()

    def tearDown(self):
        super(TestGLMTasks, self).tearDown()
        self.conf.reset()

    def test_DNSConfiguration_execute_no_vthunder_warn(self):
        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        task_path = "a10_octavia.controller.worker.tasks.glm_tasks"
        log_message = str("No vthunder therefore dns cannot be assigned.")
        expected_log = ["WARNING:{}:{}".format(task_path, log_message)]
        with self.assertLogs(task_path, level='WARN') as cm:
            dns_task.execute(None)
            self.assertEqual(expected_log, cm.output)

    def test_DNSConfiguration_execute_no_network_id_warn(self):
        vthunder = copy.deepcopy(VTHUNDER)
        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        task_path = "a10_octavia.controller.worker.tasks.glm_tasks"
        log_message = str("No networks were configured therefore "
                          "nameservers cannot be set on the "
                          "vThunder-Amphora {}").format(a10constants.MOCK_VTHUNDER_ID)
        expected_log = ["WARNING:{}:{}".format(task_path, log_message)]
        with self.assertLogs(task_path, level='WARN') as cm:
            dns_task.execute(vthunder)
            self.assertEqual(expected_log, cm.output)

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_execute_no_dns(self, network_driver_mock):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id)
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        network_driver_mock.get_network.return_value = dns_net
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        dns_task.execute(vthunder)
        self.client_mock.dns.set.assert_not_called()

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_execute_use_license_net_primary_only(self, network_driver_mock):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id, primary_dns=PRIMARY_DNS)
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        network_driver_mock.get_network.return_value = dns_net
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        dns_task.execute(vthunder)
        args, kwargs = self.client_mock.dns.set.call_args
        self.assertEqual(args, (PRIMARY_DNS, None))

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_execute_with_primary_secondary(self, network_driver_mock):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id,
                         primary_dns=PRIMARY_DNS, secondary_dns=SECONDARY_DNS)
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        network_driver_mock.get_network.return_value = dns_net
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        dns_task.execute(vthunder)
        args, kwargs = self.client_mock.dns.set.call_args
        self.assertEqual(args, (PRIMARY_DNS, SECONDARY_DNS))

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_execute_use_network_dns(self, network_driver_mock):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id)
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        dns_subnet['dns_nameservers'] = [PRIMARY_DNS, SECONDARY_DNS]
        network_driver_mock.get_network.return_value = dns_net
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        dns_task.execute(vthunder)
        args, kwargs = self.client_mock.dns.set.call_args
        self.assertEqual(args, (PRIMARY_DNS, SECONDARY_DNS))

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_execute_too_many_network_dns_warn(self, network_driver_mock):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id)
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        dns_subnet['dns_nameservers'] = [PRIMARY_DNS, SECONDARY_DNS, '3.3.3.3']
        network_driver_mock.get_network.return_value = dns_net
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock

        task_path = "a10_octavia.controller.worker.tasks.glm_tasks"
        log_message = ("More than one DNS nameserver detected on subnet {}. "
                       "Using {} as primary and {} as secondary.".format(
                           DNS_SUBNET.id, PRIMARY_DNS, SECONDARY_DNS))
        expected_log = ["WARNING:{}:{}".format(task_path, log_message)]
        with self.assertLogs(task_path, level='WARN') as cm:
            dns_task.execute(vthunder)
            self.assertEqual(expected_log, cm.output)

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_execute_use_amp_mgmt_net(self, network_driver_mock):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         primary_dns=PRIMARY_DNS, secondary_dns=SECONDARY_DNS)
        self.conf.config(group=a10constants.A10_CONTROLLER_WORKER_CONF_SECTION,
                         amp_mgmt_network=DNS_NETWORK.id)
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        network_driver_mock.get_network.return_value = dns_net
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        dns_task.execute(vthunder)
        args, kwargs = self.client_mock.dns.set.call_args
        self.assertEqual(args, (PRIMARY_DNS, SECONDARY_DNS))

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_execute_use_first_amp_boot_net(self, network_driver_mock):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         primary_dns=PRIMARY_DNS, secondary_dns=SECONDARY_DNS)
        self.conf.config(group=a10constants.A10_CONTROLLER_WORKER_CONF_SECTION,
                         amp_boot_network_list=[DNS_NETWORK.id, 'random-net'])
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        network_driver_mock.get_network.return_value = dns_net
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        dns_task.execute(vthunder)
        args, kwargs = self.client_mock.dns.set.call_args
        self.assertEqual(args, (PRIMARY_DNS, SECONDARY_DNS))

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_execute_config_precedence(self, network_driver_mock):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id,
                         primary_dns=PRIMARY_DNS, secondary_dns=SECONDARY_DNS)
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        dns_subnet['dns_nameservers'] = ['8.8.8.8', '8.8.4.4']
        network_driver_mock.get_network.return_value = dns_net
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        dns_task.execute(vthunder)
        args, kwargs = self.client_mock.dns.set.call_args
        self.assertEqual(args, (PRIMARY_DNS, SECONDARY_DNS))

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_execute_flavor_dns(self, network_driver_mock):
        flavor = {'dns': {'primary-dns': PRIMARY_DNS, 'secondary-dns': SECONDARY_DNS}}
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id)
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        network_driver_mock.get_network.return_value = dns_net
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        dns_task.execute(vthunder, flavor)
        args, kwargs = self.client_mock.dns.set.call_args
        self.assertEqual(args, (PRIMARY_DNS, SECONDARY_DNS))

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_execute_flavor_dns_precedence(self, network_driver_mock):
        flavor = {'dns': {'primary-dns': PRIMARY_DNS, 'secondary-dns': SECONDARY_DNS}}
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id,
                         primary_dns='1.0.1.0', secondary_dns='0.1.0.1')
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        dns_subnet['dns_nameservers'] = ['8.8.8.8', '8.8.4.4']
        network_driver_mock.get_network.return_value = dns_net
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        dns_task.execute(vthunder, flavor)
        args, kwargs = self.client_mock.dns.set.call_args
        self.assertEqual(args, (PRIMARY_DNS, SECONDARY_DNS))

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_execute_with_secondary_fail(self, network_driver_mock):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id,
                         secondary_dns=SECONDARY_DNS)
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        network_driver_mock.get_network.return_value = dns_net
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        self.assertRaises(a10_ex.PrimaryDNSMissing, dns_task.execute, vthunder)

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_revert_delete_dns(self, network_driver_mock):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id,
                         secondary_dns=SECONDARY_DNS)
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        network_driver_mock.get_network.return_value = dns_net
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        dns_task.revert(vthunder)
        args, kwargs = self.client_mock.dns.delete.call_args
        self.assertEqual(args, (None, SECONDARY_DNS))

    @mock.patch('a10_octavia.controller.worker.tasks.glm_tasks.DNSConfiguration.network_driver')
    def test_DNSConfiguration_revert_no_dns_return(self, network_driver_mock):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id)
        vthunder = copy.deepcopy(VTHUNDER)
        dns_net = copy.deepcopy(DNS_NETWORK)
        network_driver_mock.get_network.return_value = dns_net
        dns_subnet = copy.deepcopy(DNS_SUBNET).to_dict()
        network_driver_mock.show_subnet_detailed.return_value = {'subnet': dns_subnet}

        dns_task = task.DNSConfiguration()
        dns_task.axapi_client = self.client_mock
        dns_task.execute(vthunder)
        self.client_mock.dns.delete.assert_not_called()

    def test_ActivateFlexpoolLicense_execute_no_vthunder_warn(self):
        flexpool_task = task.ActivateFlexpoolLicense()
        flexpool_task.axapi_client = self.client_mock
        task_path = "a10_octavia.controller.worker.tasks.glm_tasks"
        log_message = str("No vthunder therefore licensing cannot occur.")
        expected_log = ["WARNING:{}:{}".format(task_path, log_message)]
        with self.assertLogs(task_path, level='WARN') as cm:
            flexpool_task.execute(None, None)
            self.assertEqual(expected_log, cm.output)

    def _template_glm_call(self):
        expected_call = {
            'token': a10constants.MOCK_FLEXPOOL_TOKEN,
            'enable_requests': True,
            'interval': None,
            'port': 443,
            'allocate_bandwidth': None,
            'use_mgmt_port': False
        }
        return expected_call

    def test_ActivateFlexpoolLicense_execute_success(self):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id,
                         flexpool_token=a10constants.MOCK_FLEXPOOL_TOKEN)
        vthunder = copy.deepcopy(VTHUNDER)
        amphora = copy.deepcopy(AMPHORA)
        interfaces = {
            'interface': {
                'ethernet-list': []
            }
        }
        expected_call = self._template_glm_call()

        flexpool_task = task.ActivateFlexpoolLicense()
        flexpool_task.axapi_client = self.client_mock
        flexpool_task.axapi_client.interface.get_list.return_value = interfaces

        flexpool_task.execute(vthunder, amphora)
        args, kwargs = self.client_mock.glm.create.call_args
        self.assertEqual(kwargs, expected_call)

    def test_ActivateFlexpoolLicense_execute_use_mgmt_port(self):
        self.conf.config(group=a10constants.A10_CONTROLLER_WORKER_CONF_SECTION,
                         amp_mgmt_network=DNS_NETWORK.id)
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id,
                         flexpool_token=a10constants.MOCK_FLEXPOOL_TOKEN)
        vthunder = copy.deepcopy(VTHUNDER)
        amphora = copy.deepcopy(AMPHORA)
        interfaces = {
            'interface': {
                'ethernet-list': []
            }
        }
        expected_call = self._template_glm_call()
        expected_call['use_mgmt_port'] = True

        flexpool_task = task.ActivateFlexpoolLicense()
        flexpool_task.axapi_client = self.client_mock
        flexpool_task.axapi_client.interface.get_list.return_value = interfaces

        flexpool_task.execute(vthunder, amphora)
        args, kwargs = self.client_mock.glm.create.call_args
        self.assertEqual(kwargs, expected_call)

    def test_ActivateFlexpoolLicense_execute_iface_up(self):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id,
                         flexpool_token=a10constants.MOCK_FLEXPOOL_TOKEN)
        vthunder = copy.deepcopy(VTHUNDER)
        amphora = copy.deepcopy(AMPHORA)
        interfaces = {
            'interface': {
                'ethernet-list': [{
                    'ifnum': 2,
                    'action': 'disable'
                }]
            }
        }

        flexpool_task = task.ActivateFlexpoolLicense()
        flexpool_task.axapi_client = self.client_mock
        flexpool_task.axapi_client.interface.get_list.return_value = interfaces

        flexpool_task.execute(vthunder, amphora)
        self.client_mock.system.action.setInterface.assert_called_with(2)

    def test_ActivateFlexpoolLicense_execute_enable_burst(self):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id,
                         flexpool_token=a10constants.MOCK_FLEXPOOL_TOKEN,
                         burst=True)
        vthunder = copy.deepcopy(VTHUNDER)
        amphora = copy.deepcopy(AMPHORA)
        interfaces = {
            'interface': {
                'ethernet-list': []
            }
        }
        expected_call = self._template_glm_call()

        flexpool_task = task.ActivateFlexpoolLicense()
        flexpool_task.axapi_client = self.client_mock
        flexpool_task.axapi_client.interface.get_list.return_value = interfaces

        flexpool_task.execute(vthunder, amphora)
        args, kwargs = self.client_mock.glm.create.call_args
        self.assertEqual(kwargs, expected_call)
        flexpool_task.axapi_client.glm.update.assert_called_with(burst=True)

    def test_ActiveFlexpoolLicense_execute_fail_LicenseOptionNotAllowed_burst(self):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         amp_license_network=DNS_NETWORK.id,
                         flexpool_token=a10constants.MOCK_FLEXPOOL_TOKEN)
        vthunder = copy.deepcopy(VTHUNDER)
        amphora = copy.deepcopy(AMPHORA)
        interfaces = {
            'interface': {
                'ethernet-list': []
            }
        }

        flexpool_task = task.ActivateFlexpoolLicense()
        flexpool_task.axapi_client = self.client_mock
        flexpool_task.axapi_client.interface.get_list.return_value = interfaces
        flexpool_task.axapi_client.glm.create.side_effect = acos_errors.LicenseOptionNotAllowed()

        flexpool_task.axapi_client = self.client_mock
        task_path = "a10_octavia.controller.worker.tasks.glm_tasks"
        log_message = ("A specified configuration option is "
                       "incompatible with license type provided. "
                       "This error can occur when the license request has "
                       "failed due to connection issue. Please check your "
                       "configured GLM network and dns settings.")
        expected_log = ["ERROR:{}:{}".format(task_path, log_message)]
        with self.assertLogs(task_path, level='ERROR') as cm:
            try:
                flexpool_task.execute(vthunder, amphora)
            except Exception:
                pass
            self.assertEqual(expected_log, cm.output)

    def test_ActivateFlexpoolLicense_revert_deactivate_license(self):
        vthunder = copy.deepcopy(VTHUNDER)
        amphora = copy.deepcopy(AMPHORA)
        flexpool_task = task.ActivateFlexpoolLicense()
        flexpool_task.axapi_client = self.client_mock

        flexpool_task.revert(vthunder, amphora)
        self.client_mock.delete.glm_license.post.assert_called()

    def test_RevokeFlexpoolLicense_execute_success(self):
        vthunder = copy.deepcopy(VTHUNDER)
        revoke_task = task.RevokeFlexpoolLicense()
        revoke_task.axapi_client = self.client_mock

        revoke_task.execute(vthunder)
        self.client_mock.delete.glm_license.post.assert_called()

    def test_RevokeFlexpoolLicense_execute_no_vthunder_warn(self):
        revoke_task = task.RevokeFlexpoolLicense()
        revoke_task.axapi_client = self.client_mock

        task_path = "a10_octavia.controller.worker.tasks.glm_tasks"
        log_message = str("No vthunder therefore license revocation cannot occur.")
        expected_log = ["WARNING:{}:{}".format(task_path, log_message)]
        with self.assertLogs(task_path, level='WARN') as cm:
            revoke_task.execute(None)
            self.assertEqual(expected_log, cm.output)

    def test_ConfigureForwardProxyServer_execute_success(self):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         proxy_host=PROXY_HOST, proxy_port=PROXY_PORT,
                         proxy_username=PROXY_USERNAME, proxy_password=PROXY_PASSWORD,
                         proxy_secret_string=PROXY_PASSWORD_VALUE)
        vthunder = copy.deepcopy(VTHUNDER)
        proxy_server = task.ConfigureForwardProxyServer()
        proxy_server.axapi_client = self.client_mock
        proxy_server.execute(vthunder)
        self.client_mock.glm.proxy_server.create.assert_called()

    def test_ConfigureForwardProxyServer_execute_flavor_success(self):
        flavor = {
            'glm-proxy-server': {
                'proxy_host': PROXY_HOST,
                'proxy_port': PROXY_PORT,
                'proxy_username': PROXY_USERNAME,
                'proxy_password': PROXY_PASSWORD,
                'proxy_secret_string': PROXY_PASSWORD_VALUE
            }
        }
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         proxy_host="10.10.10.11", proxy_port=8888,
                         proxy_username='configuser', proxy_password=False,
                         proxy_secret_string='configpwrd')
        vthunder = copy.deepcopy(VTHUNDER)
        proxy_server = task.ConfigureForwardProxyServer()
        proxy_server.axapi_client = self.client_mock
        proxy_server.execute(vthunder, flavor)
        self.client_mock.glm.proxy_server.create.assert_called_with(**{'host': PROXY_HOST,
                                                                       'port': PROXY_PORT,
                                                                       'username': PROXY_USERNAME,
                                                                       'password': PROXY_PASSWORD,
                                                                       'secret_string':
                                                                           PROXY_PASSWORD_VALUE})

    def test_ConfigureForwardProxyServer_execute_no_vthunder_warn(self):
        proxy_server = task.ConfigureForwardProxyServer()
        proxy_server.axapi_client = self.client_mock
        task_path = "a10_octavia.controller.worker.tasks.glm_tasks"
        log_message = str("No vthunder therefore forward proxy server cannot be configured.")
        expected_log = ["WARNING:{}:{}".format(task_path, log_message)]
        with self.assertLogs(task_path, level='WARN') as cm:
            proxy_server.execute(None, None)
            self.assertEqual(expected_log, cm.output)

    def test_ConfigureForwardProxyServer_execute_no_proxy_conf(self):
        self.conf.config(group=a10constants.GLM_LICENSE_CONFIG_SECTION,
                         proxy_host=None, proxy_port=None)
        vthunder = copy.deepcopy(VTHUNDER)
        proxy_server = task.ConfigureForwardProxyServer()
        proxy_server.axapi_client = self.client_mock
        proxy_server.execute(vthunder)
        self.client_mock.glm.proxy_server.create.assert_not_called()



"""add acos version column to vthunders table

Revision ID: 34a0d811941d
Revises: aa1984742932
Create Date: 2021-03-01 05:09:40.157396

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '34a0d811941d'
down_revision = 'aa1984742932'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('vthunders', sa.Column('acos_version', sa.String(36), nullable=True))


def downgrade():
    op.drop_column('vthunders', 'acos_version')
